import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    read_test_data,
    run_and_report,
)
from nef_pipelines.tools.chains.clone import clone

runner = CliRunner()
app = typer.Typer()
app.command()(clone)

INPUT_3AA_NEF = read_test_data("3aa.nef", __file__)


# noinspection PyUnusedLocal
def test_clone_basic():

    result = run_and_report(app, ["A", "2"], input=INPUT_3AA_NEF)

    EXPECTED = """\
            save_nef_molecular_system
               _nef_molecular_system.sf_category   nef_molecular_system
               _nef_molecular_system.sf_framecode  nef_molecular_system

               loop_
                  _nef_sequence.index
                  _nef_sequence.chain_code
                  _nef_sequence.sequence_code
                  _nef_sequence.residue_name
                  _nef_sequence.linking
                  _nef_sequence.residue_variant
                  _nef_sequence.cis_peptide

                 1   A   1   ALA   start    .   .
                 2   A   2   ALA   middle   .   .
                 3   A   3   ALA   end      .   .
                 4   B   1   ALA   start    .   .
                 5   B   2   ALA   middle   .   .
                 6   B   3   ALA   end      .   .
                 7   C   1   ALA   start    .   .
                 8   C   2   ALA   middle   .   .
                 9   C   3   ALA   end      .   .

               stop_

            save_
    """

    assert_lines_match(EXPECTED, isolate_frame(result.stdout, "nef_molecular_system"))


# noinspection PyUnusedLocal
def test_bad_count():

    result = run_and_report(app, ["A", "0"], input=INPUT_3AA_NEF, expected_exit_code=1)

    assert result.exit_code == 1

    assert "clone count must be > 0" in result.stdout


# noinspection PyUnusedLocal
def test_bad_target():

    result = run_and_report(app, ["B", "1"], input=INPUT_3AA_NEF, expected_exit_code=1)

    assert "couldn't find target chain B" in result.stdout


# noinspection PyUnusedLocal
def test_custom_chains():

    result = run_and_report(app, ["A", "2", "--chains", "D,E"], input=INPUT_3AA_NEF)

    EXPECTED = """\
            save_nef_molecular_system
               _nef_molecular_system.sf_category   nef_molecular_system
               _nef_molecular_system.sf_framecode  nef_molecular_system

               loop_
                  _nef_sequence.index
                  _nef_sequence.chain_code
                  _nef_sequence.sequence_code
                  _nef_sequence.residue_name
                  _nef_sequence.linking
                  _nef_sequence.residue_variant
                  _nef_sequence.cis_peptide

                 1   A   1   ALA   start    .   .
                 2   A   2   ALA   middle   .   .
                 3   A   3   ALA   end      .   .
                 4   D   1   ALA   start    .   .
                 5   D   2   ALA   middle   .   .
                 6   D   3   ALA   end      .   .
                 7   E   1   ALA   start    .   .
                 8   E   2   ALA   middle   .   .
                 9   E   3   ALA   end      .   .

               stop_

            save_
    """

    assert_lines_match(EXPECTED, isolate_frame(result.stdout, "nef_molecular_system"))


# noinspection PyUnusedLocal
def test_clone_chain_clash():

    INPUT = read_test_data("3aa_x3.nef", __file__)

    result = run_and_report(app, ["A", "2"], input=INPUT)

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    EXPECTED = """\
            save_nef_molecular_system
               _nef_molecular_system.sf_category   nef_molecular_system
               _nef_molecular_system.sf_framecode  nef_molecular_system

               loop_
                  _nef_sequence.index
                  _nef_sequence.chain_code
                  _nef_sequence.sequence_code
                  _nef_sequence.residue_name
                  _nef_sequence.linking
                  _nef_sequence.residue_variant
                  _nef_sequence.cis_peptide

                 1    A   1   ALA   start    .   .
                 2    A   2   ALA   middle   .   .
                 3    A   3   ALA   end      .   .
                 4    B   1   ALA   start    .   .
                 5    B   2   ALA   middle   .   .
                 6    B   3   ALA   end      .   .
                 7    C   1   ALA   start    .   .
                 8    C   2   ALA   middle   .   .
                 9    C   3   ALA   end      .   .
                 10   D   1   ALA   start    .   .
                 11   D   2   ALA   middle   .   .
                 12   D   3   ALA   end      .   .
                 13   E   1   ALA   start    .   .
                 14   E   2   ALA   middle   .   .
                 15   E   3   ALA   end      .   .

               stop_

            save_
    """

    assert_lines_match(EXPECTED, isolate_frame(result.stdout, "nef_molecular_system"))



import sys
from math import sqrt
from typing import Dict, List

import typer
from pynmrstar import Saveframe
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    loop_row_dict_iter,
    read_entry_from_stdin_or_exit,
    select_frames_by_name,
)
from nef_pipelines.lib.structures import AtomLabel, NewPeak, SequenceResidue, ShiftData
from nef_pipelines.lib.util import exit_error
from nef_pipelines.tools.peaks import peaks_app


@peaks_app.command()
def match(
    # ignore_amide_shifts: bool = typer.Option(False, help='ignore amides when matching'),
    # amide_weight: float = typer.Option(1.0, help='weight for the amide shifts'),
    # amide_search_region: Tuple[float,float] = typer.Option((0.2, 0.2*7), help='region to search for amide shifts in
    # the order H N'),
    names: List[str] = typer.Argument(
        ..., help="pairs of shift frame names for the chemical shifts to compare"
    ),
):
    """- match one list of peaks against another [alpha]"""

    print(
        "*** WARNING *** this command [shifts make_peaks] is only lightly tested use at your own risk!",
        file=sys.stderr,
    )

    if len(names) != 2:
        exit_error("i currently need pairs of peaks")

    entry = read_entry_from_stdin_or_exit()

    name_1, name_2 = names

    frame_1 = _select_single_frame_or_exit(entry, name_1)
    frames_2 = _select_single_frame_or_exit(entry, name_2)

    # amide_search_region = {'H': amide_search_region[0], 'N': amide_search_region[1]}

    match_peaks(frame_1, frames_2)


def _nef_frames_to_peak_shifts(frame: Saveframe) -> Dict[str, Dict[str, float]]:
    loop = frame.get_loop("nef_peak")

    indices = [
        int(tag.split("_")[-1]) for tag in loop.tags if tag.startswith("atom_name")
    ]

    peak_atom_shifts = {}
    for row in loop_row_dict_iter(loop):
        key = row["peak_id"]

        atom_shifts = {}
        peak_atom_shifts[key] = atom_shifts
        for index in indices:
            shift = row[f"position_{index}"]
            atom_name = row[f"atom_name_{index}"]

            atom_shifts[atom_name] = shift

    return peak_atom_shifts


def _find_best_match(
    shifts: Dict[str, float],
    match_shifts: Dict[str, Dict[str, float]],
    weights=Dict[str, float],
):

    weight_values_1 = _get_weighted_terms(shifts, weights)

    result = {}

    for peak_id, match_shifts in match_shifts.items():
        weight_values_2 = _get_weighted_terms(match_shifts, weights)

        len_values_1 = len(weight_values_1)
        len_values_2 = len(weight_values_2)

        if len_values_1 != len_values_2:
            continue

        names_1 = set(weight_values_1.keys())
        names_2 = set(weight_values_2.keys())

        if names_1 != names_2:
            continue

        axis_distances = []

        for atom_name in names_1:
            axis_distances.append(
                (weight_values_1[atom_name] - weight_values_2[atom_name]) ** 2
            )
        distance = sqrt(sum(axis_distances))

        result.setdefault(distance, []).append(peak_id)

    return result
    #     distances = []
    #     value_2
    #     distances.append(value_1**2)


def _get_weighted_terms(shifts, weights):

    result = {}
    for name, value in shifts.items():
        weight_key = name[0]
        weight = weights[weight_key]
        result[name] = value / weight
    return result


def _nef_frames_to_peak_by_id(frame):
    loop = frame.get_loop("nef_peak")

    indices = [
        int(tag.split("_")[-1]) for tag in loop.tags if tag.startswith("atom_name")
    ]

    peak_by_id = {}
    for row in loop_row_dict_iter(loop):
        peak_id = row["peak_id"]

        shifts = []
        for index in indices:
            chain_code = row[f"chain_code_{index}"]
            sequence_code = row[f"sequence_code_{index}"]
            residue_name = row[f"residue_name_{index}"]
            atom_name = row[f"atom_name_{index}"]
            position = row[f"position_{index}"]

            residue = SequenceResidue(chain_code, sequence_code, residue_name)
            atom = AtomLabel(residue, atom_name)

            shifts.append(ShiftData(atom, position))

        peak_by_id[peak_id] = NewPeak(shifts, peak_id)

    return peak_by_id


def _get_assignment_tuple(peak: NewPeak):
    result = []
    for shift in peak.shifts:
        residue = shift.atom.residue
        result.extend(
            (
                residue.chain_code,
                residue.sequence_code,
                residue.residue_name,
                shift.atom.atom_name,
            )
        )

    return tuple(result)


def _get_assignment_string(peak: NewPeak):
    result = []
    for shift in peak.shifts:
        residue = shift.atom.residue
        result.append(
            f"#{residue.chain_code}:{residue.sequence_code}[{residue.residue_name}]@{shift.atom.atom_name}"
        )

    return "-".join(result)


def match_peaks(peak_frame_1: Saveframe, peak_frame_2: Saveframe):

    num_dimensions = int(peak_frame_1.get_tag("num_dimensions")[0])

    shifts_1 = _nef_frames_to_peak_shifts(peak_frame_1)
    shifts_2 = _nef_frames_to_peak_shifts(peak_frame_2)

    peaks_by_id_1 = _nef_frames_to_peak_by_id(peak_frame_1)
    peaks_by_id_2 = _nef_frames_to_peak_by_id(peak_frame_2)

    info_by_assignment = {}
    no_matches = []
    for peak_1, target_shifts in shifts_1.items():
        matches = _find_best_match(
            target_shifts, shifts_2, weights={"H": 1.0, "N": 7.0, "C": 2.0, ".": 1.0}
        )

        if len(matches) > 0:
            best_match = sorted(matches)[0]
            peak_2 = matches[best_match]
            peak_2_str = [str(match) for match in matches[best_match]]
            info_string = [
                peak_1,
                ", ".join(peak_2_str),
                len(peak_2),
                best_match,
                "",
                *_get_assignment_tuple(peaks_by_id_1[peak_1]),
                "",
                *_get_assignment_tuple(peaks_by_id_2[peak_2[0]]),
            ]
            assignment = _get_assignment_tuple(peaks_by_id_2[peak_2[0]])

            info_by_assignment[assignment] = info_string
        else:
            no_matches.append(peak_1)

    table = []
    for key in sorted(info_by_assignment):
        table.append((info_by_assignment[key]))

    # TODO: move to peaks table
    # TODO: colorise matches?
    # TODO: add better sorting

    headings = "f1.pk f2.pks num dist".split()
    for peak in 1, 2:
        headings.append(f"f{peak}:")
        for dim in range(1, num_dimensions + 1):
            headings.extend(f"chn-{dim} seq-{dim} resn-{dim} atm-{dim}".split())

    print(f"frame 1: {peak_frame_1.name}")
    print(f"frame 2: {peak_frame_2.name}")
    print()
    if len(no_matches) > 0:
        no_matches = ", ".join([str(peak) for peak in no_matches])
        print(f"note no matches for the following peak-1's: {no_matches}")
        print()

    table.sort(key=lambda x: x[0])
    print(tabulate(table, tablefmt="plain", headers=headings))

    # residue_atom_shifts_1 = _peak_list_to_peak_shifts(shifts_1)
    # residue_atom_shifts_2 = _peak_list_to_peak_shifts(shifts_2)

    # table = []
    #
    # for (chain, residue), shifts in residue_atom_shifts_1.items():
    #
    #     ignored_atoms = set('H N'.split()) if ignore_amides else set()
    #     matches = _find_best_match(shifts, residue_atom_shifts_2, ignored_atoms, amide_weight, amide_search_region)
    #
    #     ordered_matches = []
    #     for key in sorted(matches.keys()):
    #         name = f'{matches[key][0]}{matches[key][1]}'
    #         value = f'{key:.3}'
    #
    #         ordered_matches.append(f'{name} {value}')
    #
    #     row = [f'{chain}{residue}', *ordered_matches[:5]]
    #     table.append(row)
    #
    # lengths = []
    # for row in table:
    #     for column in row[1:]:
    #         lengths.append(len(column))
    #
    # max_column_length = max(lengths) + 1
    #
    # for row in table:
    #     for i, column in enumerate(row[1:], start=1):
    #         column_length = len(column)
    #         column_fields = column.split()
    #         pad_length = max_column_length - column_length
    #         pad = ' ' * pad_length
    #         column = pad.join(column_fields)
    #         row[i] = column
    #
    #
    # print(tabulate(table, tablefmt='plain'))


def _shift_list_to_residue_atom_shifts(shifts_1):
    residue_shifts = {}

    for shift in shifts_1:
        chain_code = shift.atom.residue.chain_code
        sequence_code = shift.atom.residue.sequence_code
        atom_name = shift.atom.atom_name
        if "-" in sequence_code:
            sequence_code, neg_offset = sequence_code.split("-")
            atom_name = f"{atom_name}-{neg_offset}"

        key = chain_code, sequence_code
        shift_value = shift.value

        residue_shifts.setdefault(key, {})[atom_name] = shift_value

    return residue_shifts


def _select_single_frame_or_exit(entry, name):

    frames = select_frames_by_name(entry.frame_list, name)

    if len(frames) == 0:
        exit_error(
            f"no frames selected from entry: {entry.entry_id} by selector {name}"
        )

    if len(frames) > 1:

        msg = f"""\
            multiple frames selected from entry: {entry.entry_id} by selector {name}
            frames selected were {' '.join([frame.name for frame in frames])}
        """
        exit_error(msg)

        exit_error(msg)

    return frames[0]


# def _find_best_match(shifts_1, residue_atom_shifts_2, ignored_atoms, amide_weight, amide_search_region):
#     matches = {}
#     for key, shifts_2 in residue_atom_shifts_2.items():
#         atom_names_1 = set(shifts_1.keys())
#         atom_names_2 = set(shifts_2.keys())
#
#         common_atom_names = atom_names_1.intersection(atom_names_2)
#
#
#
#         if len(common_atom_names) != 0:
#
#             differences = {}
#             for atom_name in common_atom_names:
#                 diff = abs(shifts_1[atom_name] - shifts_2[atom_name])
#                 if atom_name in 'H N'.split():
#                     diff = diff * amide_weight
#                 differences[atom_name] = diff
#
#             ok = True
#
#             for amide_name in amide_search_region:
#                 if amide_name in differences:
#                     if differences[amide_name] > amide_search_region[amide_name]:
#                         ok = False
#
#             if ok:
#                 for name in ignored_atoms:
#                     if name in matches:
#                         del matches[name]
#
#                 matches[sum(differences.values())] = key
#
#     return matches
#
#
# def _squeeze_spaces_left(string):
#     pre_len = len(string)
#     string = string.replace(' ', '')
#     post_len = len(string)
#     pad = " " * (pre_len - post_len)
#     string = f'{pad}{string}'
#
#     return string
#
#
#



from pathlib import Path
from typing import List

import typer
from ordered_set import OrderedSet
from typer import Argument, Option

from nef_pipelines.lib.nef_lib import (
    SELECTORS_LOWER,
    SelectionType,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.util import STDIN
from nef_pipelines.tools.chains import chains_app

app = typer.Typer()

# TODO: it would be nice to put the chains with the first molecular system frame


# noinspection PyUnusedLocal
@chains_app.command()
def rename(
    new_old: List[str] = Argument(
        ...,
        help="old chain-code followed by new chain-code, multiple pairs of old and new "
        "chain-codes can be provided",
    ),
    # TODO: just use --category instead to make life simpler???
    selector_type: SelectionType = typer.Option(
        SelectionType.ANY,
        "-t",
        "--selector-type",
        help=f"how to select frames to renumber, can be one of: {SELECTORS_LOWER}."
        "Any will match on names first and then if there is no match attempt to match on category",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    frame_selectors: List[str] = Option(
        None,
        "-f",
        "--frame",
        help="limit changes to a a particular frame by a selector which can be a frame name or category, "
        "note: wildcards [*] are allowed. Frames are selected by name and subsequently by category if the name "
        "doesn't match [-t /--selector-type allows you to force which selection type to use]. If no frame"
        "names or categories are provided chain-codes are renamed in all frames.",
    ),
):
    """- change the name of chains across one or multiple frames"""
    old = new_old[0]
    new = new_old[1]

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    changes = 0
    changed_frames = OrderedSet()

    frames_to_process = select_frames(entry, frame_selectors, selector_type)

    for save_frame in frames_to_process:
        for loop in save_frame.loop_iterator():
            for tag in loop.get_tag_names():
                tag_parts = tag.split(".")
                if tag_parts[-1].startswith("chain_code"):
                    tag_values = loop[tag]
                    for i, row in enumerate(tag_values):
                        if row == old:
                            tag_values[i] = new
                            changes += 1
                            changed_frames.add(save_frame.name)

                    loop[tag] = tag_values

    print(entry)



import typer

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.xeasy.importers.peaks import peaks

app = typer.Typer()
app.command()(peaks)


EXPECTED = """
save_nef_nmr_spectrum_xeasy_basic                                  # noqa: E501
   _nef_nmr_spectrum.sf_category                nef_nmr_spectrum
   _nef_nmr_spectrum.sf_framecode               nef_nmr_spectrum_xeasy_basic
   _nef_nmr_spectrum.num_dimensions             3
   _nef_nmr_spectrum.chemical_shift_list        .
   _nef_nmr_spectrum.experiment_classification  .
   _nef_nmr_spectrum.experiment_type            .

   loop_
      _nef_spectrum_dimension.dimension_id
      _nef_spectrum_dimension.axis_unit
      _nef_spectrum_dimension.axis_code
      _nef_spectrum_dimension.spectrometer_frequency
      _nef_spectrum_dimension.spectral_width
      _nef_spectrum_dimension.value_first_point
      _nef_spectrum_dimension.folding
      _nef_spectrum_dimension.absolute_peak_positions
      _nef_spectrum_dimension.is_acquisition

     1   ppm   1H    600.123   0.001   8.732     none   true   false
     2   ppm   1H    600.123   4.984   8.950     none   true   false
     3   ppm   15N   60.833    0.023   115.276   none   true   true

   stop_

   loop_
      _nef_spectrum_dimension_transfer.dimension_1
      _nef_spectrum_dimension_transfer.dimension_2
      _nef_spectrum_dimension_transfer.transfer_type
      _nef_spectrum_dimension_transfer.is_indirect

     1   2   onebond   false
     2   3   onebond   false

   stop_

   loop_
      _nef_peak.index
      _nef_peak.peak_id
      _nef_peak.chain_code_1
      _nef_peak.sequence_code_1
      _nef_peak.residue_name_1
      _nef_peak.atom_name_1
      _nef_peak.chain_code_2
      _nef_peak.sequence_code_2
      _nef_peak.residue_name_2
      _nef_peak.atom_name_2
      _nef_peak.chain_code_3
      _nef_peak.sequence_code_3
      _nef_peak.residue_name_3
      _nef_peak.atom_name_3
      _nef_peak.position_1
      _nef_peak.position_uncertainty_1
      _nef_peak.position_2
      _nef_peak.position_uncertainty_2
      _nef_peak.position_3
      _nef_peak.position_uncertainty_3
      _nef_peak.height
      _nef_peak.height_uncertainty
      _nef_peak.volume
      _nef_peak.volume_uncertainty
      _nef_peak.ccpn_comment

     1   1   A   2   MET   H   A   3   ARG   H     A   4   GLN   N   8.731   .   8.723   .   115.265   .   .   .   5.5   0.0   'MAP    403'
     2   2   .   .   .     .   .   .   .     .     .   .   .     .   8.731   .   4.61    .   115.275   .   .   .   5.5   0.0   .
     3   3   A   5   THR   H   A   6   MET   HA2   A   7   LEU   N   8.732   .   4.192   .   115.254   .   .   .   5.5   0.0   'MAP    404'
     4   3   A   5   THR   H   A   6   MET   HA3   A   7   LEU   N   8.732   .   4.192   .   115.254   .   .   .   5.5   0.0   'MAP    405'

   stop_

save_
""".replace(
    NOQA_E501, ""
)


def test_basic():

    data_sequence = read_test_data("basic_sequence.nef", __file__)
    peaks_path = path_in_test_data(__file__, "basic.peaks")

    result = run_and_report(app, [peaks_path], input=data_sequence)

    loop = isolate_frame(result.stdout, "nef_nmr_spectrum_xeasy_basic")

    assert_lines_match(EXPECTED, str(loop))



import typer

from nef_pipelines import nef_app

series_app = typer.Typer()


if nef_app.app:
    nef_app.app.add_typer(
        series_app, name="series", help="- carry out operations on a data series"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.tools.series.build  # noqa: F401
    import nef_pipelines.tools.series.table  # noqa: F401



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.fasta.exporters.sequence import sequence

runner = CliRunner()
app = typer.Typer()
app.command()(sequence)

EXPECTED_3AA = """
>test NEFPLS | CHAIN: A | START: 1

AAA
"""


# noinspection PyUnusedLocal
def test_3aa():

    path = path_in_test_data(__file__, "3aa.nef")
    result = run_and_report(app, ["--in", path, "-"])

    assert_lines_match(result.stdout, EXPECTED_3AA)


EXPECTED_3AA_2_CHAINS = """
>test NEFPLS | CHAIN: A | START: 1

AAA
>test NEFPLS | CHAIN: B | START: 1

CCC
"""


def test_3aa_2_chains():

    path = path_in_test_data(__file__, "3aa_2_chains.nef")
    result = run_and_report(app, ["--in", path, "-"])

    assert_lines_match(result.stdout, EXPECTED_3AA_2_CHAINS)


EXPECTED_THX = """
>test NEFPLS | CHAIN: thx | START: 1
AAA
"""


def test_petes_thx_chain():
    path = path_in_test_data(__file__, "3aa_thx.nef")
    result = run_and_report(app, ["--in", path, "-"])

    assert_lines_match(result.stdout, EXPECTED_THX)



import pytest
import typer
from pynmrstar import Entry

from nef_pipelines.lib.globals_lib import debug_clear_globals
from nef_pipelines.lib.test_lib import (
    assert_frame_category_exists,
    read_test_data,
    run_and_report,
)
from nef_pipelines.tools.globals import globals_

app = typer.Typer()
app.command()(globals_)


@pytest.fixture
def setup_globals():
    debug_clear_globals()
    yield
    debug_clear_globals()


def test_null(setup_globals):

    result = run_and_report(app, [])

    assert result.stdout.strip() == ""


def test_basic(setup_globals):
    data = read_test_data(
        "header.nef",
        __file__,
    )

    result = run_and_report(app, ["--verbose"], input=data)

    entry = Entry.from_string(result.stdout)

    assert_frame_category_exists(entry, "nefpls_globals")

    globals_frames = entry.get_saveframes_by_category("nefpls_globals")

    globals_frame = globals_frames[0]
    assert globals_frame.get_tag("_nefpls_globals.verbose")[0] == "True"
    assert globals_frame.get_tag("_nefpls_globals.force")[0] == "False"
    assert "global options" in globals_frame.get_tag("_nefpls_globals.info")[0]


def test_passthru(setup_globals):
    data = read_test_data(
        "header.nef",
        __file__,
    )

    result = run_and_report(app, [], input=data)

    entry = Entry.from_string(result.stdout)

    assert_frame_category_exists(entry, "nefpls_globals")



# TODO: tests on output to multiple files
#       better default formatting of headings for table formats that can support spaces
#       headers/dividers if mutiple frames are to be output into the same file / stream

import csv
import re
import sys
from collections import Counter
from enum import Enum, auto
from pathlib import Path
from typing import Dict, List

import typer
import wcmatch.fnmatch as fnmatch
from fyeah import f
from pynmrstar import Saveframe
from tabulate import tabulate as tabulate_formatter

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import (
    STDOUT,
    exit_error,
    is_int,
    parse_comma_separated_options,
)
from nef_pipelines.tools.frames import frames_app

NO_FLAGS = 0x0000
ALL_LOOPS = "*"

UNDERSCORE = "_"

parser = None

EXIT_ERROR = 1

FORMAT_TO_EXTENSION = {
    "csv": "csv",
    "fancy_grid": "txt",
    "fancy_outline": "txt",
    "github": "md",
    "grid": "txt",
    "html": "html",
    "latex": "tex",
    "latex_booktabs": "tex",
    "latex_longtable": "tex",
    "latex_raw": "tex",
    "mediawiki": "txt",
    "moinmoin": "txt",
    "orgtbl": "org",
    "pipe": "txt",
    "plain": "txt",
    "presto": "presto",
    "pretty": "txt",
    "psql": "mkd",
    "rst": "rst",
    "simple": "txt",
    "textile": "textile",
    "tsv": "tsv",
    "unsafehtml": "html",
}

formats = list(FORMAT_TO_EXTENSION)
FORMAT_HELP = f"""
    format for the table, [possible formats are: {", ".join(formats)}: most of those provided by tabulate, if you
    require other formats provided by tabulate contact the authors]
"""
FRAME_AND_LOOP_SELECTOR_HELP = """
Selectors for frames and loops to tabulate. Multiple frames and loops can be selected. Frame names and loop categories /
indices are separated by a dot [e.g. frame-name.loop-category or frame-name.loop-index]. Wild cards are allowed and are
attached to the start and end of the frame-names and loop-categories by default when searching. Specific loops can be
chosen by index or loop category. Exact matching of frame names and loop categories can be made using the --exact
option. Details of frames in a file loops can be listed using 'nef frame list -vvv'
"""

ABBREVIATED_HEADINGS = {
    "index": "ind",
    "chain_code": "chain",
    "sequence_code": "seq",
    "residue_name": "resn",
    "linking": "link",
    "cis_peptide": "cis",
    "ccpn_compound_name": "ccpn_name",
    "ccpn_chain_role": "ccpn_role",
    "ccpn_comment": "comment",
    "ccpn_dataset_serial": "data_serial",
    "program_name": "name",
    "program_version": "version",
    "script_name": "script",
    "peak_id": "id",
    "volume": "vol",
    "volume_uncertainty": "vol-err",
    "height_uncertainty": "height-err",
    "position": "pos",
    "position_uncertainty": "pos-err",
    "ccpn_figure_of_merit": "merit",
    "ccpn_annotation": "ann",
    "ccpn_peak_list_serial": "ccpn-serial",
    "atom-name": "atom",
}
SELECT_HELP = """
    a list of columns to select for output, to specify multiple columns specify the option multiple times or use a comma
    separated list. Prepend columns to exclude with - and columns to include with a +. If neither a + or - is prepended
    to a column a + is assumed, a -- at the start of a column name is treated as an escape for a - at the start of a
    columns name. Wildcards are applied before and after the selections if the --exact option is not specified.
"""

OUT_HELP = """
    file or files to write to including templates, the place holders {entry} {frame} and {loop} will get replaced the
    current entry id frame name and loop category, if multiple frame names and loop names apply multiple files maybe
    output
"""

ABBREVIATE_HELP = """
    abbreviate column headings. Also convert _ to - in  formats wich require single word headings such as csv and tsv
"""


class ColumnSelectionType(Enum):
    INCLUDE = auto()
    EXCLUDE = auto()


# noinspection PyUnusedLocal
@frames_app.command(no_args_is_help=True)
def tabulate(
    input_file: Path = typer.Option(
        None,
        "--in",
        metavar="|PIPE|",
        help="pipe to read NEF data from, for testing [overrides stdin !use stdin instead!]",
    ),
    out_format: str = typer.Option(
        "plain", "-f", "--format", help=FORMAT_HELP, metavar="format"
    ),
    verbose: bool = typer.Option(
        False, "-v", "--verbose", help="include frame names etc in output"
    ),
    full: bool = typer.Option(
        False, "--full", help="don't suppress empty columns [default: False]"
    ),
    exact: bool = typer.Option(
        False,
        "--exact",
        help="when matching frames and categories do it exactly",
    ),
    no_title: bool = typer.Option(
        False, "--no-title", help="don't display the frame name as a title"
    ),
    select_columns: List[str] = typer.Option(["+*"], help=SELECT_HELP),
    out: str = typer.Option(str(STDOUT), help=OUT_HELP),
    abbreviate: bool = typer.Option(False, "--abbreviate", help=ABBREVIATE_HELP),
    frame_loop_selectors: List[str] = typer.Argument(
        None, help=FRAME_AND_LOOP_SELECTOR_HELP, metavar="<frame-and-loop-selectors>..."
    ),
):
    """
    - tabulate loops from frames in a NEF file. notes: 1. using the name of a frame as a selector will tabulate all
      in the frame while using frame.loop_index [e.g. molecular_system.1] or frame.loop_name [e.g. shift_frame_HN.peaks]
      will tabulate a specific loop. 2. wild cards can be used for frame names e.g. mol*tem would select
      molecular_system and any other frame whose name contains mol followed by sys and is case-insensitive 3. all loop
      and frame selector components are surrounded by *'s before use unless the --exact option is used 3. by default
      empty columns are ignored unless the option --full is used"""

    args = get_args()

    if not args.frame_loop_selectors:
        args.frame_loop_selectors = [
            ALL_LOOPS,
        ]

    args.select_columns = _build_column_selections(args.select_columns, args.exact)

    entry = read_or_create_entry_exit_error_on_bad_file(args.input_file)

    tabulate_frames(entry, args)


def _build_column_selections(column_selections, exact):
    column_selections = parse_comma_separated_options(column_selections)

    curated_column_selections = []

    SIMPLE_SELECTIONS = {
        "": (ColumnSelectionType.INCLUDE, "*"),
        "*": (ColumnSelectionType.INCLUDE, "*"),
        "-": (ColumnSelectionType.EXCLUDE, "*"),
        "+": (ColumnSelectionType.INCLUDE, "*"),
        "-*": (ColumnSelectionType.EXCLUDE, "*"),
        "+*": (ColumnSelectionType.INCLUDE, "*"),
    }
    for column_selection in column_selections:
        if column_selection in SIMPLE_SELECTIONS:
            curated_column_selections.append(SIMPLE_SELECTIONS[column_selection])
            continue
        if column_selection.startswith("--"):
            curated_column_selections.append(
                (ColumnSelectionType.INCLUDE, column_selection)
            )
            continue
        if column_selection.startswith("+"):
            curated_column_selections.append(
                (ColumnSelectionType.INCLUDE, column_selection[1:])
            )
            continue
        if column_selection.startswith("-"):
            curated_column_selections.append(
                (ColumnSelectionType.EXCLUDE, column_selection[1:])
            )
            continue
        if not column_selection[0] in ("-", "+"):
            curated_column_selections.append(
                (ColumnSelectionType.INCLUDE, column_selection)
            )
            continue

    expanded_curated_selections = []
    if not exact:
        for action, selection in curated_column_selections:
            if len(selection) > 1 and not selection.endswith("*"):
                selection = f"{selection}*"
            if len(selection) > 1 and not selection.startswith("*"):
                selection = f"*{selection}"
            expanded_curated_selections.append((action, selection))
        curated_column_selections = expanded_curated_selections

    return curated_column_selections


def tabulate_frames(entry, args):
    frames_to_tabulate = _get_frames_to_tabulate(
        entry, args.frame_loop_selectors, args.exact
    )
    frame_indices = _select_loop_indices(args.frame_loop_selectors)

    seen_files = set()
    for frame_name, frame_data in frames_to_tabulate.items():

        frame_category = frame_data.category

        category_length = len(frame_category)

        frame_id = frame_data.name[category_length:].lstrip("_")
        frame_id = frame_id.strip()
        frame_id = frame_id if len(frame_id) > 0 else ""

        for i, loop in enumerate(frame_data.loops, start=1):
            if _should_output_loop(
                i, frame_name, frame_indices, loop.category, args.exact
            ):
                _output_loop(
                    loop, frame_id, frame_category, entry.entry_id, args, seen_files
                )


def _select_loop_indices(frames):
    frame_indices = {}
    for frame_selector in frames:
        frame_selector_index = frame_selector.split(".")
        if len(frame_selector_index) == 1:
            frame_name_selector = frame_selector_index[0]
            frame_indices.setdefault(frame_name_selector, set()).add(ALL_LOOPS)
        elif len(frame_selector_index) == 2:
            frame_name_selector = frame_selector_index[0]
            index = frame_selector_index[1]
            index = int(index) if is_int(index) else index
            frame_indices.setdefault(frame_name_selector, set()).add(index)
        else:
            exit_error(
                f'[frames tabulate] too many fields int the frame selector {".".join(frame_selector)}'
            )
    for frame_selector, indices in frame_indices.items():
        if ALL_LOOPS in indices and len(indices) > 1:
            indices.remove(ALL_LOOPS)
            index_strings = ", ".join([str(index) for index in indices])
            message = f"""
                [frames tabulate] incompatible selections, you selected all loops and a specific index for
                frame: {frame_selector} indices were: {index_strings}
                """

            exit_error(message)

    return frame_indices


def _remove_empty_columns(tabulation, headers):

    counter = Counter()
    for row in tabulation:
        for i, header in enumerate(headers):
            if row[i] != UNUSED:
                counter[i] += 1

    empty_columns = []
    for column_index, heading in enumerate(headers):
        if not counter[column_index]:
            empty_columns.append(column_index)

    empty_columns = list(reversed(empty_columns))

    for row in tabulation:
        for column_index in empty_columns:
            del row[column_index]

    for column_index in empty_columns:
        del headers[column_index]

    return tabulation, headers


class OutputFormat(Enum):
    PER_ENTRY = auto()
    PER_FRAME = auto()
    PER_LOOP = auto()
    STDOUT = auto()


# from bing chatgpt
def _replace_quoted_string(string):
    return re.sub(r"`([^`]+)`", r"::\1", string)


def _output_loop(loop_data, frame_id, frame_category, entry_id, args, seen_files):

    out_type = _get_out_type_or_exit_error(args.out)

    entry = entry_id  # noqa F841
    frame = f"{frame_category}_{frame_id}"  # noqa F841
    loop = loop_data.category.lstrip("_")  # noqa F841
    file_extension = FORMAT_TO_EXTENSION[args.out_format]
    expanded_file_name = f(args.out)
    out_name = f"{expanded_file_name}.{file_extension}"
    out_name = _replace_quoted_string(out_name)

    if out_type == OutputFormat.STDOUT:
        out_file = sys.stdout
    else:
        append = out_name in seen_files
        if not append:
            out_file = open(out_name, "wt")
            seen_files.add(out_name)
        else:
            out_file = open(out_name, "at")

    if args.verbose:
        print()
        if frame_id:
            print(f"{frame_id}: {frame_category}/{loop_data.category[1:]}")
        else:
            print(f"{frame_category}/{loop_data.category[1:]}")
        print()
    table = []

    headers = loop_data.tags
    used_headers = _get_selected_columns(headers, args.select_columns, args.exact)

    if used_headers and loop_data.data:
        for line in loop_data.data:
            row = list(line)
            out_row = []
            for column_index, (column, header) in enumerate(zip(row, headers)):
                if header in used_headers:
                    out_row.append(column)
            table.append(out_row)

        if not args.full:
            table, used_headers = _remove_empty_columns(table, used_headers)

        frame_category = loop_data.category.lstrip("_")
        if args.abbreviate:
            used_headers = _abbreviate_headers(used_headers, ABBREVIATED_HEADINGS)

        if not args.no_title:
            print(frame_category)
            print("-" * len(frame_category))
            print()

        if args.out_format in ["csv", ""]:
            writer = csv.writer(out_file, lineterminator="\n")
            used_headers = [header.strip() for header in used_headers]
            writer.writerow(used_headers)
            for row in table:
                writer.writerow(row)
        else:
            print(
                tabulate_formatter(
                    table, headers=used_headers, tablefmt=args.out_format
                ),
                file=out_file,
            )

        if out_type != OutputFormat.STDOUT:
            out_file.close()


def _get_selected_columns(headers, columns_selections, exact):

    selected_columns = set(headers)
    match_flags = fnmatch.IGNORECASE if not exact else NO_FLAGS

    for selection_type, selection_string in columns_selections:
        current_selections = set(
            [
                header
                for header in headers
                if fnmatch.fnmatch(header, selection_string, flags=match_flags)
            ]
        )

        if selection_type == ColumnSelectionType.INCLUDE:
            selected_columns.update(current_selections)
        else:
            selected_columns.difference_update(current_selections)

    # this ensures the columns are in the order in the file
    result = []
    for header in headers:
        if header in selected_columns:
            result.append(header)

    return result


def _get_out_type_or_exit_error(out_string):

    out_string = str(out_string)
    have_entry = "{entry}" in out_string
    have_frame = "{frame}" in out_string
    have_frame_id = "{frame_id}" in out_string
    have_frame_category = "{frame_category}" in out_string
    if have_frame_id and have_frame_id:
        have_frame = True
    have_loop = "{loop}" in out_string

    if have_loop and not have_frame:
        msg = f"""
            if template contains {{loop}} it must also contain either {{frame}}...
            the output template was {out_string}
        """
        exit_error(msg)

    if (have_frame_id and not have_frame_category) or (
        have_frame_category and not have_frame_id
    ):
        msg = f"""
            if you have frame_id you must have frame_category and vice versa
            the output template was {out_string}
        """
        exit_error(msg)

    output_type = OutputFormat.PER_ENTRY

    if have_frame:
        output_type = OutputFormat.PER_FRAME
    if have_loop:
        output_type = OutputFormat.PER_LOOP
    if have_entry:
        output_type = OutputFormat.PER_ENTRY

    if out_string == "-":
        output_type = OutputFormat.STDOUT

    return output_type


def _abbreviate_headers(headers: List[str], abbreviations: Dict[str, str]) -> List[str]:

    result = []

    for header in headers:
        for stem in abbreviations:

            if header.startswith(stem):

                header = f"{abbreviations[stem]}{header[len(stem):]} "

        result.append(header.replace("_", "-"))

    return result


def _should_output_loop(index, frame_name, frame_indices, loop_category, exact):

    do_output = False
    match_flags = fnmatch.IGNORECASE if not exact else NO_FLAGS
    for frame_selector in frame_indices:
        if exact:
            matched = frame_name == frame_selector
        else:
            matched = fnmatch.fnmatch(
                frame_name.lstrip("_"), f"*{frame_selector}*", flags=match_flags
            )

        if matched:
            indices = frame_indices[frame_selector]

            if index in indices:
                do_output = True

            if ALL_LOOPS in indices:
                do_output = True

            if exact:
                output_tests = [
                    loop_category.lstrip("_") == target_index
                    for target_index in indices
                ]
            else:
                output_tests = [
                    fnmatch.fnmatch(
                        loop_category, f"*{target_index}*", flags=match_flags
                    )
                    for target_index in indices
                ]

            if any(output_tests):
                do_output = True

            if do_output:
                break

    return do_output


def _count_loop_rows(save_frame: Saveframe) -> int:
    count = 0
    for loop in save_frame.loops:
        count += len(loop)

    return count


def _get_frames_to_tabulate(entry, frame_selectors, exact):

    frames_with_empty_loops = _get_loops_with_empty_frames_and_errors(entry)

    frames_to_tabulate = _select_chosen_frames(entry, frame_selectors, exact)

    _remove_empty_frames_and_warn(frames_to_tabulate, frames_with_empty_loops)

    return frames_to_tabulate


def _remove_empty_frames_and_warn(frames_to_tabulate, frames_with_empty_loops):
    for frame_name in frames_to_tabulate:
        if frame_name in frames_with_empty_loops:
            print(frames_with_empty_loops[frame_name], file=sys.stderr)
    for frame_name in frames_with_empty_loops:
        if frame_name in frames_to_tabulate:
            del frames_to_tabulate[frame_name]


def _select_chosen_frames(entry, frame_selectors, exact):
    frames_to_tabulate = {}
    match_flags = fnmatch.IGNORECASE if not exact else NO_FLAGS
    for frame_name, data in entry.frame_dict.items():
        if len(frame_selectors) > 0:
            for frame_selector in frame_selectors:
                frame_selector = frame_selector.split(".")[0]

                if fnmatch.fnmatch(
                    frame_name, f"*{frame_selector}*", flags=match_flags
                ):
                    frames_to_tabulate[frame_name] = data
        else:

            frames_to_tabulate[frame_name] = data
    return frames_to_tabulate


def _get_loops_with_empty_frames_and_errors(entry):
    frames_with_empty_loops = {}
    for frame_name, data in entry.frame_dict.items():
        if len(data.loops) == 0:
            msg = f"WARNING: frame {frame_name} has no loops and the frame will be ignored"
            frames_with_empty_loops[frame_name] = msg

        if frame_name not in frames_with_empty_loops and _count_loop_rows(data) == 0:
            msg = f"WARNING: frame {frame_name} has loops but they contain no data and the frame will be ignored"
            frames_with_empty_loops[frame_name] = msg
    return frames_with_empty_loops



BMRB_URL_TEMPLATE = "https://bmrb.io/ftp/pub/bmrb/entry_directories/bmr{entry_number}/bmr{entry_number}_3.str"

# TODO these should be in a resource file?
SHORTCUTS = {
    "UBIQUITIN": "bmr5387",
    "GB1": "bmr7280",
    "GB3": "bmr25910",
    "MBP": "bmr4354",
    "MSG": "bmr5471",
    "INSULIN": "bmr1000",
}

SHORTCUT_ENTRYS = {name: id[3:] for name, id in SHORTCUTS.items()}
SHORTCUT_URLS = {
    name: BMRB_URL_TEMPLATE.format(entry_number=entry_number)
    for name, entry_number in SHORTCUT_ENTRYS.items()
}
SHORTCUT_NAMES = ", ".join({name.lower(): name for name in SHORTCUTS.keys()})



import os
from pathlib import Path

import typer
from pynmrstar import Entry

from nef_pipelines.lib.test_lib import (
    assert_frame_category_exists,
    assert_lines_match,
    read_test_data,
    run_and_report,
)
from nef_pipelines.tools.save import save

app = typer.Typer()
app.command()(save)

EXPECTED = """
-------------------------------------------------- xplor --------------------------------------------------
data_xplor

save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   AAAA   1    ALA   start    .   .
     2   AAAA   2    ALA   middle   .   .
     3   AAAA   3    ALA   end      .   .
     4   BBBB   11   ALA   start    .   .
     5   BBBB   12   ALA   middle   .   .
     6   BBBB   13   ALA   end      .   .

   stop_

save_

-------------------------------------------------- test --------------------------------------------------
data_test

save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide
      _nef_sequence.ccpn_comment
      _nef_sequence.ccpn_chain_role
      _nef_sequence.ccpn_compound_name
      _nef_sequence.ccpn_chain_comment

     1   A   3   HIS   .   .   .   .   .   Sec5   .
     2   A   4   MET   .   .   .   .   .   Sec5   .
     3   B   5   ARG   .   .   .   .   .   Sec5   .
     4   B   6   GLN   .   .   .   .   .   Sec5   .
     5   C   7   PRO   .   .   .   .   .   Sec5   .

   stop_

save_

----------------------------------------------------------------------------------------------------------
"""


EXPECTED_XPLOR = """
data_xplor

save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   AAAA   1    ALA   start    .   .
     2   AAAA   2    ALA   middle   .   .
     3   AAAA   3    ALA   end      .   .
     4   BBBB   11   ALA   start    .   .
     5   BBBB   12   ALA   middle   .   .
     6   BBBB   13   ALA   end      .   .

   stop_

save_
"""

EXPECTED_TEST = """
data_test

save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide
      _nef_sequence.ccpn_comment
      _nef_sequence.ccpn_chain_role
      _nef_sequence.ccpn_compound_name
      _nef_sequence.ccpn_chain_comment

     1   A   3   HIS   .   .   .   .   .   Sec5   .
     2   A   4   MET   .   .   .   .   .   Sec5   .
     3   B   5   ARG   .   .   .   .   .   Sec5   .
     4   B   6   GLN   .   .   .   .   .   Sec5   .
     5   C   7   PRO   .   .   .   .   .   Sec5   .

   stop_

save_
"""


# noinspection PyUnusedLocal
def test_save_multi_stream_stdout():
    data = read_test_data(
        "multi.nef",
        __file__,
    )
    result = run_and_report(app, ["-"], input=data)

    assert_lines_match(EXPECTED, result.stdout)


def test_save_multi_stream_stdout_single_file():
    data = read_test_data(
        "multi.nef",
        __file__,
    )
    result = run_and_report(app, "--single-file -".split(), input=data)

    assert_lines_match(EXPECTED, result.stdout)


EXPECTED_NO_HEADER = "\n".join(
    [line for line in EXPECTED.splitlines() if not line.startswith("--")]
)


def test_save_multi_stream_stdout_no_header():
    data = read_test_data(
        "multi.nef",
        __file__,
    )
    result = run_and_report(app, "--no-header -".split(), input=data)

    assert_lines_match(EXPECTED_NO_HEADER, result.stdout)


def test_save_multi_stream_to_directory(tmp_path):
    data = read_test_data(
        "multi.nef",
        __file__,
    )

    result = run_and_report(
        app,
        [
            str(tmp_path),
        ],
        input=data,
    )

    assert_lines_match(result.stdout, data)

    assert Path(tmp_path / "xplor.nef").exists()
    assert Path(tmp_path / "test.nef").exists()

    assert_lines_match(EXPECTED_XPLOR, Path(tmp_path / "xplor.nef").read_text())
    assert_lines_match(EXPECTED_TEST, Path(tmp_path / "test.nef").read_text())


def test_save_multi_stream_to_single_file(tmp_path):
    data = read_test_data(
        "multi.nef",
        __file__,
    )

    test_file_path = tmp_path / "test_all.nef"
    result = run_and_report(
        app,
        [
            "--single-file",
            str(test_file_path),
        ],
        input=data,
    )

    assert_lines_match(result.stdout, data)

    assert Path(tmp_path / "test_all.nef").exists()

    assert_lines_match(EXPECTED, Path(tmp_path / "test_all.nef").read_text())


def test_save_multi_stream_to_directory_files_exist(tmp_path):
    data = read_test_data(
        "multi.nef",
        __file__,
    )

    (tmp_path / "xplor.nef").write_text("existing file")

    result = run_and_report(
        app,
        [
            str(tmp_path),
        ],
        input=data,
        expected_exit_code=1,
    )

    assert "when trying to write the file" in result.stdout
    assert "it already exists" in result.stdout


def test_save_multi_stream_to_directory_files_exist_force(tmp_path):
    data = read_test_data(
        "multi.nef",
        __file__,
    )

    (tmp_path / "xplor.nef").write_text("existing file")

    result = run_and_report(
        app,
        [
            "--force",
            str(tmp_path),
        ],
        input=data,
    )

    assert_lines_match(result.stdout, data)

    assert Path(tmp_path / "xplor.nef").exists()
    assert Path(tmp_path / "test.nef").exists()

    assert_lines_match(EXPECTED_XPLOR, Path(tmp_path / "xplor.nef").read_text())
    assert_lines_match(EXPECTED_TEST, Path(tmp_path / "test.nef").read_text())


def test_save_multi_stream_to_files(tmp_path):
    data = read_test_data(
        "multi.nef",
        __file__,
    )

    test_path_1 = tmp_path / "test_1.nef"
    test_path_2 = tmp_path / "test_2.nef"

    result = run_and_report(
        app, f"{str(test_path_1)} {str(test_path_2)}".split(), input=data
    )

    assert_lines_match(result.stdout, data)

    assert test_path_1.exists()
    assert test_path_2.exists()

    assert_lines_match(EXPECTED_XPLOR, test_path_1.read_text())
    assert_lines_match(EXPECTED_TEST, test_path_2.read_text())


def test_save_multi_stream_to_too_few_files(tmp_path):
    data = read_test_data(
        "multi.nef",
        __file__,
    )

    test_path_1 = tmp_path / "test_1.nef"

    result = run_and_report(
        app, f"{str(test_path_1)}".split(), input=data, expected_exit_code=1
    )

    assert test_path_1.exists()
    assert_lines_match(EXPECTED_XPLOR, test_path_1.read_text())

    assert "Number of entries does not match number of file paths" in result.stdout
    assert "i ran out of file paths to write to" in result.stdout


def test_save_no_stream(tmp_path):

    test_path_1 = tmp_path / "test_1.nef"

    result = run_and_report(app, f"{str(test_path_1)}".split(), expected_exit_code=1)

    assert (
        "you must provide at least one entry to write, did you input a NEF file stream?"
        in result.stdout
    )


def test_save_multi_stream_to_no_files(tmp_path):
    data = read_test_data(
        "multi.nef",
        __file__,
    )

    os.chdir(tmp_path)

    result = run_and_report(app, [], input=data, expected_exit_code=0)

    assert_lines_match(result.stdout, data)

    assert Path(tmp_path / "xplor.nef").exists()
    assert Path(tmp_path / "test.nef").exists()

    assert_lines_match(EXPECTED_XPLOR, Path(tmp_path / "xplor.nef").read_text())
    assert_lines_match(EXPECTED_TEST, Path(tmp_path / "test.nef").read_text())


def test_save_single_file_to_directory_error(tmp_path):

    result = run_and_report(
        app, f"--single-file {str(tmp_path)}".split(), expected_exit_code=1
    )

    assert (
        "you must provide at least one entry to write, did you input a NEF file stream?"
        in result.stdout
    )


def test_save_single_file_to_multipe_files_error(tmp_path):

    data = read_test_data(
        "multi.nef",
        __file__,
    )

    test_path_1 = tmp_path / "test_1.nef"
    test_path_2 = tmp_path / "test_2.nef"

    result = run_and_report(
        app,
        f"--single-file {str(test_path_1)} {test_path_2}".split(),
        input=data,
        expected_exit_code=1,
    )

    assert (
        "the single file option is incompatible with multiple file paths, there were 2 file paths"
        in result.stdout
    )
    assert str(test_path_1) in result.stdout
    assert str(test_path_2) in result.stdout


def test_save_single_stream_stdout():
    data = read_test_data(
        "header.nef",
        __file__,
    )

    result = run_and_report(
        app,
        ["-"],
        input=data,
    )

    assert "--------" not in result.stdout
    assert_frame_category_exists(result.stdout, "nef_nmr_meta_data")


def test_multi_line_string():
    data = read_test_data(
        "header_globals.nef",
        __file__,
    )

    result = run_and_report(
        app,
        ["-"],
        input=data,
    )

    Entry.from_string(result.stdout)
    # no exceptions is all that is needed


def test_globals_cleanup():
    data = read_test_data(
        "header_globals.nef",
        __file__,
    )

    result = run_and_report(
        app,
        ["-"],
        input=data,
    )

    assert_frame_category_exists(result.stdout, "nefpls_globals", count=0)


def test_globals_cleanup_off():
    data = read_test_data(
        "header_globals.nef",
        __file__,
    )

    result = run_and_report(
        app,
        ["--no-globals-cleanup", "-"],
        input=data,
    )

    assert_frame_category_exists(result.stdout, "nefpls_globals", count=1)



import csv
import re
from enum import auto
from io import TextIOWrapper
from pathlib import Path
from textwrap import dedent
from typing import List, Tuple

import typer
from pynmrstar import Entry, Loop, Saveframe
from strenum import StrEnum
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import (
    get_residue_name_from_lookup,
    sequence_from_entry,
    sequence_from_entry_or_exit,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.structures import AtomLabel, RdcRestraint, SequenceResidue
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    get_display_file_name,
    is_float,
    is_int,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.csv import import_app

app = typer.Typer()

HELP_FOR_FRAME_NAME = "name for the frame, note white spaces will be replace by _"
HELP_FOR_FORMATS = """\
    CSV like formats that can be read [CSV: comma separated variables, TSV tab separated variables,
    SSV space separated variables, AUTO read file and guess format from contents]
"""


class CsvLikeFormats(StrEnum):
    CSV = auto()
    TSV = auto()
    SSV = auto()
    AUTO = auto()
    csv = auto()
    tsv = auto()
    ssv = auto()
    auto = auto()


CHAIN_CODE = "chain_code"
SEQUENCE_CODE = "sequence_code"
ATOM = "atom"
VALUE = "value"
VALUE_UNCERTAINTY = "value_uncertainty"

CHAIN_CODE_1 = f"{CHAIN_CODE}_1"
CHAIN_CODE_2 = f"{CHAIN_CODE}_2"
CHAINS_1_2 = f"{CHAIN_CODE_1} {CHAIN_CODE_2}"
SEQUENCE_CODE_1 = f"{SEQUENCE_CODE}_1"
SEQUENCE_CODE_2 = f"{SEQUENCE_CODE}_2"
SEQUENCE_CODE_1_2 = f"{SEQUENCE_CODE_1} {SEQUENCE_CODE_2}"
ATOM_1 = f"{ATOM}_1"
ATOM_2 = f"{ATOM}_2"
ATOMS_1_2 = f"{ATOM_1} {ATOM_2}"

REQUIRED_HEADERS_SHORT = f"{SEQUENCE_CODE} {VALUE}".split()
REQUIRED_HEADERS_LONG = f"{CHAINS_1_2} {SEQUENCE_CODE_1_2} {ATOMS_1_2} {VALUE}".split()


COLUMN_SEPARATORS_MAY_HAVE_CHANGED = (
    "note: column separators shown here may different to those in the original file..."
)
DEFAULT_ATOMS_HELP = (
    "the atoms to use if atoms aren't defined in the file, there should be two. "
    "This option can be used mutiple twice to define the atoms or you can use a "
    "comma separated list"
)


class BadSequenceException(Exception):
    pass


def _exit_bad_sequence(e, entry_input, sequence_table=None):
    msg = (
        f"{dedent(str(e))}\n"
        f"the source of the residue sequence was: {get_display_file_name(entry_input)}\n\n"
    )
    if not sequence_table:
        msg += "note: if you would like to see a complete input sequence use the verbose option"
    else:
        msg += "sequence was:\n\n"
        msg += sequence_table

    exit_error(msg)


@import_app.command(no_args_is_help=True)
def rdcs(
    entry_input: Path = typer.Option(
        STDIN,
        metavar="<INPUT>",
        help="file to read NEF data from [stdin = -]",
    ),
    csv_format: CsvLikeFormats = typer.Option(
        CsvLikeFormats.AUTO, "-f", "--format", help=HELP_FOR_FORMATS
    ),
    csv_file_encoding: str = typer.Option(
        "utf-8-sig", "-e", "--encoding", help="encoding for the csv file"
    ),
    verbose: bool = typer.Option(False, help="report verbose information"),
    default_atoms: List[str] = typer.Option(
        [], "-a", "--atoms", help=DEFAULT_ATOMS_HELP
    ),
    default_chain_code: str = typer.Option(
        "A",
        "-c",
        "--chain-code",
        help="default chain code to use if none is provided in the file",
    ),
    csv_file: Path = typer.Argument(None, metavar="<CSV-FILE>"),
):
    default_atoms = parse_comma_separated_options(default_atoms)

    if not default_atoms:
        default_atoms = ("H", "N")

    csv_format = csv_format.upper()

    entry = read_entry_from_file_or_stdin_or_exit_error(entry_input)

    try:
        pipe(
            entry,
            default_chain_code,
            default_atoms,
            csv_file,
            csv_file_encoding,
            csv_format,
        )
    except BadSequenceException as e:

        sequence_table = _tabulate_sequence(entry) if verbose else None
        _exit_bad_sequence(e, entry_input, sequence_table)

    print(entry)


def pipe(
    entry: Entry,
    chain_code: str,
    atoms: Tuple[str, str],
    csv_file: Path,
    csv_file_encoding: str,
    csv_format: CsvLikeFormats,
) -> Entry:

    sequence = sequence_from_entry_or_exit(entry)

    lookup = sequence_to_residue_name_lookup(sequence)

    encoding = {"encoding": csv_file_encoding}
    file_rdcs = _parse_csv(
        csv_file, encoding, chain_code, atoms, lookup, csv_file, csv_format
    )

    frame = _rdcs_to_frame(file_rdcs)

    add_frames_to_entry(entry, [frame])

    return entry


# noinspection PyUnusedLocal
def _rdcs_to_frame(rdcs: List[RdcRestraint], frame_code="rdcs"):
    RDC_FRAME_CATEGORY = "nef_rdc_restraint_list"
    RDC_LOOP_CATEGORY = "nef_rdc_restraint"

    frame_code = f"{RDC_FRAME_CATEGORY}_{frame_code}"

    frame = Saveframe.from_scratch(frame_code, RDC_FRAME_CATEGORY)

    frame_tags = (
        "restraint_origin",
        "tensor_magnitude",
        "tensor_rhombicity",
        "tensor_chain_code",
        "tensor_sequence_code",
        "tensor_residue_name",
    )

    frame.add_tag("sf_category", RDC_FRAME_CATEGORY)
    frame.add_tag("sf_framecode", frame_code)

    for tag in frame_tags:
        frame.add_tag(tag, ".")

    loop = Loop.from_scratch()
    frame.add_loop(loop)

    loop_tags = (
        "index",
        "restraint_id",
        "restraint_combination_id",
        "chain_code_1",
        "sequence_code_1",
        "residue_name_1",
        "atom_name_1",
        "chain_code_2",
        "sequence_code_2",
        "residue_name_2",
        "atom_name_2",
        "weight",
        "target_value",
        "target_value_uncertainty",
        "lower_linear_limit",
        "lower_limit",
        "upper_limit",
        "upper_linear_limit",
        "scale",
        "distance_dependent",
    )

    loop.set_category(RDC_LOOP_CATEGORY)
    loop.add_tag(loop_tags)

    for index, rdc in enumerate(rdcs):
        data = {
            "index": index,
            "restraint_id": index,
            "chain_code_1": rdc.atom_1.residue.chain_code,
            "sequence_code_1": rdc.atom_1.residue.sequence_code,
            "residue_name_1": rdc.atom_1.residue.residue_name,
            "atom_name_1": rdc.atom_1.atom_name,
            "chain_code_2": rdc.atom_2.residue.chain_code,
            "sequence_code_2": rdc.atom_2.residue.sequence_code,
            "residue_name_2": rdc.atom_2.residue.residue_name,
            "atom_name_2": rdc.atom_2.atom_name,
            "weight": 1.0,
            "target_value": rdc.value,
            "target_value_uncertainty": rdc.value_uncertainty,
            "scale": 1.0,
        }
        loop.add_data([data])

    return frame


def _parse_csv(
    csv_file, encoding, default_chain_code, atoms, lookup, file_name, csv_format
):
    column_offsets = {}
    headers = [
        CHAIN_CODE,
        *REQUIRED_HEADERS_SHORT,
        *REQUIRED_HEADERS_LONG,
        VALUE_UNCERTAINTY,
    ]
    data = []
    weight = 1.0
    with open(csv_file, **encoding) as csv_fp:

        rdc_reader = _get_csv_reader_for_format(csv_format, csv_fp, encoding)

        for i, row in enumerate(rdc_reader):

            if i == 0:
                header_row = [elem.lower() for elem in row]

                _exit_if_headers_dont_look_right(row, i + 1, header_row, file_name)

                for header in headers:
                    if header in header_row:
                        column_offsets[header] = header_row.index(header)

            else:

                if CHAIN_CODE_1 in column_offsets and CHAIN_CODE_2 in column_offsets:
                    chain_code_1 = row[column_offsets[CHAIN_CODE_1]]
                    chain_code_2 = row[column_offsets[CHAIN_CODE_2]]
                else:
                    chain_code_1 = default_chain_code
                    chain_code_2 = default_chain_code

                if (
                    SEQUENCE_CODE_1 in column_offsets
                    and SEQUENCE_CODE_2 in column_offsets
                ):
                    sequence_code_1 = row[column_offsets[SEQUENCE_CODE_1]]
                    sequence_code_2 = row[column_offsets[SEQUENCE_CODE_2]]
                else:
                    sequence_code_1 = row[column_offsets[SEQUENCE_CODE]]
                    sequence_code_2 = row[column_offsets[SEQUENCE_CODE]]

                if SEQUENCE_CODE in column_offsets:
                    msg = f"for this converter to work the {SEQUENCE_CODE} must be an integer"
                else:
                    msg = "for this converter to work the the sequence codes must be integers"

                msg_1 = msg + f", {SEQUENCE_CODE_1} wasn't"
                msg_2 = msg + f", {SEQUENCE_CODE_2} wasn't"
                _exit_error_if_value_not_type(
                    sequence_code_1, int, msg_1, row, i + 1, file_name
                )
                _exit_error_if_value_not_type(
                    sequence_code_2, int, msg_2, row, i + 1, file_name
                )

                sequence_code_1 = int(sequence_code_1)
                sequence_code_2 = int(sequence_code_2)

                value = row[column_offsets[VALUE]]
                if not is_float(value):
                    msg = f"{VALUE}s must be floats but the input wasn't"
                    _exit_error_if_value_not_type(
                        value, float, msg, row, i + 1, file_name
                    )
                value = float(value)

                if VALUE_UNCERTAINTY in column_offsets:
                    value_uncertainty = row[column_offsets[VALUE_UNCERTAINTY]]
                    if not is_float(value_uncertainty):
                        msg = (
                            f"{VALUE_UNCERTAINTY} smust be floats but the input wasn't"
                        )
                        _exit_error_if_value_not_type(
                            value_uncertainty, float, msg, row, i + 1, file_name
                        )
                    value_uncertainty = float(value_uncertainty)
                else:
                    value_uncertainty = UNUSED

                residue_name_1 = _lookup_residue_name_or_exit(
                    chain_code_1, sequence_code_1, lookup
                )
                residue_name_2 = _lookup_residue_name_or_exit(
                    chain_code_2, sequence_code_2, lookup
                )

                residue_1 = SequenceResidue(
                    chain_code_1, sequence_code_1, residue_name_1
                )
                residue_2 = SequenceResidue(
                    chain_code_2, sequence_code_2, residue_name_2
                )

                if ATOM_1 in column_offsets and ATOM_2 in column_offsets:
                    atom_1 = AtomLabel(residue_1, row[column_offsets[ATOM_1]])
                    atom_2 = AtomLabel(residue_2, row[column_offsets[ATOM_2]])
                else:
                    atom_1 = AtomLabel(residue_1, atoms[0])
                    atom_2 = AtomLabel(residue_2, atoms[1])

                rdc = RdcRestraint(
                    atom_1, atom_2, value, value_uncertainty, weight=weight
                )

                data.append(rdc)

    return data


def _exit_if_headers_dont_look_right(row: List[str], row_number, header_row, file_name):

    if "value_uncertainty" in row:
        row = list(row)
        row.remove("value_uncertainty")

    ok = False
    if len(row) == len(REQUIRED_HEADERS_SHORT):
        if len(header_row) < len(REQUIRED_HEADERS_SHORT) and len(
            set(header_row).intersection(set(REQUIRED_HEADERS_SHORT))
        ) != len(REQUIRED_HEADERS_SHORT):
            msg = f"""\
                the column headers don't look right they should include {', '.join(REQUIRED_HEADERS_SHORT)}
                i have {', '.join(header_row)} at line {row_number} value of line was
                {', '.join(row)} [{COLUMN_SEPARATORS_MAY_HAVE_CHANGED}]
                in file {file_name}
            """
            exit_error(msg)
        else:
            ok = True

    if (
        not ok
        and len(row) == len(REQUIRED_HEADERS_LONG)
        and len(set(header_row).intersection(set(REQUIRED_HEADERS_LONG)))
        != len(REQUIRED_HEADERS_LONG)
    ):
        msg = f"""\
            the column headers don't look right they should include {', '.join(REQUIRED_HEADERS_LONG)}
            i have {', '.join(header_row)} at line {row_number} value of line was
            {', '.join(row)} [{COLUMN_SEPARATORS_MAY_HAVE_CHANGED}]
            in file {file_name}
        """
        exit_error(msg)


def _exit_error_if_value_not_type(value, type, value_msg, row, row_number, file_name):
    TYPES = {int: (is_int, "integer"), float: (is_float, "float")}

    converter, type_name = TYPES[type]

    if not converter(value):
        msg = f"""\
            at line {row_number} in file {file_name} for the value {value}
            {value_msg}
            {', '.join(row)} [{COLUMN_SEPARATORS_MAY_HAVE_CHANGED}]
        """
        exit_error(msg)


def _get_csv_reader_for_format(csv_format, csv_fp, encoding):
    if csv_format == CsvLikeFormats.AUTO:
        dialect = csv.Sniffer().sniff(csv_fp.read(1024))
        csv_fp.seek(0)
        rdc_reader = csv.reader(csv_fp, dialect)
    elif csv_format == CsvLikeFormats.TSV:
        rdc_reader = csv.reader(csv_fp, delimiter="\t")
    elif csv_format == CsvLikeFormats.CSV:
        rdc_reader = csv.reader(csv_fp)
    else:
        lines = []
        for line in csv_fp:
            line = re.sub(r"\s+", "\t", line)
            lines.append(line.strip())
        lines = "\n".join(lines)
        csv_fp = TextIOWrapper(lines, encoding)
        rdc_reader = csv.reader(csv_fp, delimiter="\t")
    return rdc_reader


def _lookup_residue_name_or_exit(chain_code, sequence_code, lookup):
    residue_name = get_residue_name_from_lookup(chain_code, sequence_code, lookup)
    if residue_name == UNUSED:
        msg = f"""
                            There is no residue with chain {chain_code} and sequence code {sequence_code} in the input
                            sequence so I can't find a residue name for this residue
                        """
        raise BadSequenceException(msg)

    return residue_name


def _tabulate_sequence(entry: Entry) -> None:

    sequence = sequence_from_entry(entry)
    table = []
    for residue in sequence:
        row = [residue.chain_code, residue.sequence_code, residue.residue_name]
        table.append(row)

    if not table:
        result = "NO SEQUENCE..!"
    else:
        headers = "chain_code sequence_code residue_name".split()
        headers = [header.replace("_", " ") for header in headers]
        result = tabulate(table, headers=headers)

    return result



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    read_test_data,
    run_and_report,
)
from nef_pipelines.tools.chains.list import list as list_app

app = typer.Typer()
app.command()(list_app)


# noinspection PyUnusedLocal
def test_frame_basic():

    input_data = read_test_data("multi_chain.nef", __file__)

    result = run_and_report(app, [], input=input_data)

    EXPECTED = "A B C"

    assert_lines_match(EXPECTED, result.stdout)



import contextlib
import functools
import inspect
import io
import os
import sys
import traceback
import warnings
from argparse import Namespace
from enum import auto
from fnmatch import fnmatch
from math import floor
from pathlib import Path
from textwrap import dedent
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    TextIO,
    Tuple,
    TypeVar,
    Union,
)

import click
from pynmrstar import Entry, Loop, Saveframe
from strenum import StrEnum
from tabulate import tabulate

from nef_pipelines.lib.constants import (
    EXIT_ERROR,
    NEF_META_DATA,
    NEF_PIPELINES,
    NEF_UNKNOWN,
)
from nef_pipelines.lib.globals_lib import get_global
from nef_pipelines.lib.header_lib import (
    create_header_frame,
    get_creation_time,
    get_uuid,
)
from nef_pipelines.lib.structures import LineInfo

UNKNOWN_INPUT_SOURCE = "unknown"

FOUR_SPACES = " " * 4

STDIN = Path("-")
STDOUT = Path("-")

NEWLINE = "\n"


def _get_loop_by_category_or_none(frame: Saveframe, category: str) -> Loop:

    result = None
    if f"_{category}" in frame.loop_dict.keys():
        result = frame.get_loop(category)

    return result


def loop_add_data_tag_dict(loop: Loop, data: Dict[str, object]) -> None:
    tagged_data = []
    for category_tag in loop.get_tag_names():
        _, tag = category_tag.split(".")
        if tag in data:
            tagged_data.append(data[tag])
        else:
            tagged_data.append(NEF_UNKNOWN)

    loop.add_data(tagged_data)


# TODO: there is no space for scrript arguments!
def fixup_metadata(entry: Entry, name: str, version: str, script: str):
    """
    add a new entry to a nef metadata frame

    Args:
        entry (Entry): the target entry
        name (str): the name of the program used
        version (str): the vserion of the program used
        script (str): the script used

    Returns:
        Entry: the modified entry
    """

    if entry is not None and NEF_META_DATA in entry.category_list:
        meta_frame = entry[NEF_META_DATA]

        last_program = meta_frame.get_tag("program_name")
        last_program = last_program[0] if last_program else "."

        last_program_version = meta_frame.get_tag("program_version")
        last_program_version = last_program_version[0] if last_program_version else "."

        last_script_name = meta_frame.get_tag("script_name")
        last_script_name = last_script_name[0] if last_script_name else "."

        meta_frame.add_tag("program_name", name, update=True)
        meta_frame.add_tag("program_version", version, update=True)
        meta_frame.add_tag("script_name", script, update=True)
        creation_date = get_creation_time()
        meta_frame.add_tag("creation_date", creation_date, update=True)
        uuid = get_uuid(NEF_PIPELINES, creation_date)
        meta_frame.add_tag("uuid", uuid, update=True)

        run_history_loop = _get_loop_by_category_or_none(meta_frame, "nef_run_history")
        if run_history_loop is not None:
            if run_history_loop.get_tag("run_number"):
                run_number_tags = run_history_loop.get_tag("run_number")
                run_numbers = [int(run_number) for run_number in run_number_tags]
                last_run_number = max(run_numbers)
                next_run_number = last_run_number + 1
            else:
                next_run_number = 1

            data = {
                "run_number": next_run_number,
                "program_name": last_program,
                "program_version": last_program_version,
                "script_name": last_script_name,
            }
            loop_add_data_tag_dict(run_history_loop, data)
        else:
            run_history_loop = Loop.from_scratch("nef_run_history")
            run_history_loop.add_tag(
                ["run_number", "program_name", "program_version", "script_name"]
            )
            run_history_loop.add_data(["."] * 4)
    else:
        header = create_header_frame(NEF_PIPELINES, get_version(), script)
        entry.add_saveframe(header)


class StringIteratorIO(io.TextIOBase):

    # https://stackoverflow.com/questions/12593576/adapt-an-iterator-to-behave-like-a-file-like-object-in-python
    # second answer with more votes!

    def __init__(self, iter):
        self._iter = iter
        self._left = ""

    def readable(self):
        return True

    def _read1(self, n=None):
        while not self._left:
            try:
                self._left = next(self._iter)
            except StopIteration:
                break
        ret = self._left[:n]
        self._left = self._left[len(ret) :]
        return ret

    def read(self, n=None):
        line = []
        if n is None or n < 0:
            while True:
                m = self._read1()
                if not m:
                    break
                line.append(m)
        else:
            while n > 0:
                m = self._read1(n)
                if not m:
                    break
                n -= len(m)
                line.append(m)
        return "".join(line)

    def readline(self):
        line = []
        while True:
            i = self._left.find("\n")
            if i == -1:
                line.append(self._left)
                try:
                    self._left = next(self._iter)
                except StopIteration:
                    self._left = ""
                    break
            else:
                line.append(self._left[: i + 1])
                self._left = self._left[i + 1 :]
                break
        return "".join(line)


# https://stackoverflow.com/questions/6657820/how-to-convert-an-iterable-to-a-stream/20260030#20260030
def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE):
    """
    Lets you use an iterable (e.g. a generator) that yields bytestrings as a read-only
    input stream.

    The stream implements Python 3's newer I/O API (available in Python 2's io module).
    For efficiency, the stream is buffered.
    """

    return StringIteratorIO(iterable)


def running_in_pycharm():
    return "PYCHARM_HOSTED" in os.environ


def get_text_from_file_or_exit(file_name: Path) -> str:
    """
    get text from stdin or from a file argument

    Args:
        file_name (str): name of the file Path('-') indicates stdin

    Returns:
        str: text from file

    """

    result = None

    if file_name == Path("-") or not file_name:
        if sys.stdin.isatty():
            exit_error("attempting to read from a terminal")
        result = sys.stdin.read()
    else:
        try:
            with open(file_name) as fh:
                result = fh.read()
        except IOError as e:
            exit_error(f"couldn't read from {file_name} because of an error", e)

    return result


def get_pipe_file(args: Namespace) -> Optional[TextIO]:
    """
    get an input on stdin or from an argument called pipe

    Args:
        args (Namespace): command line arguments

    Returns:
        TextIO: an input stream or None

    """

    result = None
    if "input" in args and args.input and args.input != STDIN:
        with open(args.input) as fh:
            pipe_lines = fh.read()
        result = io.StringIO(pipe_lines)
    # pycharm doesn't treat stdstreams correcly and hangs
    elif not sys.stdin.isatty() and not running_in_pycharm():
        result = iter(sys.stdin.read().split("\n"))

    return StringIteratorIO(result) if result else None


def get_pipe_file_text(args: Namespace) -> Optional[str]:
    """
    get an input on stdin or from an argument called pipe

    Args:
        args (Namespace): command line arguments

    Returns:
        TextIO: an input stream or None

    """

    result = None
    if "pipe" in args and args.pipe:
        result = open(args.pipe, "r").read()
    # pycharm doesn't treat stdstreams correcly and hangs
    elif not sys.stdin.isatty() and not running_in_pycharm():
        result = sys.stdin.read()

    return result if result else None


def in_pytest():
    return "pytest" in sys.modules


def script_name(file: str) -> Path:
    """
    get the name of the script

    Args:
        file (str): the name of the file

    Returns:
        Path: path to the file
    """
    try:
        current_context = click.get_current_context()
    except RuntimeError:

        current_context = Namespace(command_path="unknown unknown")

    if in_pytest():
        command = current_context.command_path.split()
    else:
        command = current_context.command_path.split()[1:]

    command = f"{'/'.join(command)}.py"

    return command


def read_from_file_or_exit(file_path: Path, target: str = UNKNOWN_INPUT_SOURCE) -> str:

    if file_path == STDIN and sys.stdin.isatty():

        msg = """
            while trying to read from STDIN [-] the input is not a stream
            [did you forget to add a nef file to your pipeline?]
        """
        exit_error(msg)

    display_file_name = get_display_file_name(file_path)

    fh = sys.stdin

    if not file_path == STDIN:
        try:
            fh = open(file_path)
        except Exception as e:
            msg = f"couldn't open {display_file_name} to read {target}"
            exit_error(msg, e)

    try:
        text = fh.read()
    except Exception as e:
        msg = f"couldn't read {target} from {display_file_name}"
        exit_error(msg, e)

    if file_path != STDIN:
        fh.close()

    return text


def _script_to_command(script: str) -> str:
    """
    turns a script path into the equivalent nef pipelines command
    :param script: the script path
    :return: the command name
    """
    script_path = script.split("/")
    script_path[-1] = script_path[-1][: -len(".py")]
    return " ".join(script_path)


def exit_error(msg, exception=None):
    """
    print an error message and exit error

    Args:
        msg: the message
    """

    # This should only happen in verbose mode
    if exception is not None:
        exc_info = sys.exc_info()
        traceback.print_exception(*exc_info, file=sys.stderr)
        print(file=sys.stderr)

    msg = dedent(msg).split("\n")

    print(file=sys.stderr)

    script = script_name(__file__)

    command = _script_to_command(script)

    print(f"ERROR [in: {command}]: {msg[0]}", file=sys.stderr)
    print(file=sys.stderr)

    for line in msg[1:]:
        print(f"  {line}", file=sys.stderr)
    print("exiting...", file=sys.stderr)
    sys.exit(EXIT_ERROR)


def process_stream_and_add_frames(
    frames: List[Saveframe], input_args: Namespace
) -> Entry:
    """
    take a set of save frames and either add them to a stream from stdin / a pipe-file or create
    a new stream with a proper NEF metadata header

    Args:
        frames: a set of save frames
        input_args: command line arguments for an entry_name and pipe file source

    Returns:
        a new entry containing the frames
    """

    try:
        stream = get_pipe_file(input_args)
    except Exception as e:
        exit_error(f"failed to load pipe file because {e}")

    if stream is not None:
        lines = stream.read()
    else:
        lines = ""

    if len(lines.strip()) == 0:
        stream = None

    new_entry = (
        Entry.from_string(lines)
        if stream
        else Entry.from_scratch(input_args.entry_name)
    )

    fixup_metadata(new_entry, NEF_PIPELINES, get_version(), script_name(__file__))

    # TODO: check if frame exists
    for frame in frames:
        new_entry.add_saveframe(frame)

    return new_entry


def is_int(value: str) -> bool:
    """
    Check if a string is an integer
    Args:
        value (str): the putative integer

    Returns bool:
        true is if its an integer

    """
    result = False
    try:
        int(value)
        result = True
    except (ValueError, TypeError):
        pass

    return result


def is_float(value: str) -> bool:
    """
    Check if a string is an float
    Args:
        value (str): the putative float

    Returns bool:
        true is if its an integer

    """
    result = False
    try:
        float(value)
        result = True
    except (ValueError, TypeError):
        pass

    return result


T = TypeVar("T")


def nef_pipelines_root():
    """
    get the root of the nef pipelines installation

    Returns:
        the root of the nef pipelines installation
    """
    return Path(__file__).parent.parent.parent


# https://stackoverflow.com/questions/312443/how-do-i-split-a-list-into-equally-sized-chunks
def chunks(input: Iterator[T], n: int) -> Iterator[List[T]]:
    """Yield successive n-sized chunks from lst.
       lst: the list to chunk
       n: the chunk size

    Returns:
        an iterator of chunks from lst of length T

    """
    lst = list(input)
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def _build_int_float_error_message(message, line_info, field, format):
    messages = []
    if message:
        messages.append(message)
    if line_info:
        message = f"""
                    couldn't convert {line_info.line} to format

                    at {line_info.line_no + 1} in f{line_info.file_name}

                    line value was

                    {line_info.line}

                    field was

                    {field}
                """
        messages.append(message)
    message = "\n\n".join(messages)
    return message


def read_float_or_exit(
    string: str, line_info: LineInfo = None, field="unknown", message=None
) -> float:
    """
    convert a string to an float or exit with error including line information

    Args:
        string: string to parse
        line_info: the line the string came from
        field: the name of the field defininbg the string (ignored if there is no line info)
        message: any other messages to show

    Returns:
        a float
    """

    format = "float"
    try:
        result = float(string)
    except Exception:
        message = _build_int_float_error_message(message, line_info, field, format)

        exit_error(message)

    return result


def read_integer_or_exit(
    string: str, line_info: LineInfo = None, field="unknown", message=None
) -> int:
    """
    convert a string to an int or exit with error including line information

    Args:
        string: string to parse
        line_info: the line the string came from
        field: the name of the field defining the string (ignored if there is no line info)
        message: any other messages to show

    Returns:
        an integer
    """

    format = "integer"
    try:
        result = int(string)
    except Exception:

        message = _build_int_float_error_message(message, line_info, field, format)

        exit_error(message)

    return result


# TODO test this!
def parse_comma_separated_options(
    lists: Union[List[List[str]], List[str], str]
) -> List[str]:
    """
    Take a mixed list of strings or strings that can be parsed as a comma separated list of strings
    and make a list of all the items

    e.g.

    test = [['1','2','3'], 'a,b,c', ['4','5']]

    gives

    ['1','2','3','a','b','c','4','5']

    :param lists: a mixture of  lists of string or comma separated as a at sting
    :return: list of items
    """
    result = []
    for item in lists:
        if isinstance(item, str):
            item = item.strip(",")
            result.extend(item.split(","))
        else:
            result.append(item)

    return result


def end_with_ordinal(n):
    return str(n) + {1: "st", 2: "nd", 3: "rd"}.get(
        4 if 10 <= n % 100 < 20 else n % 10, "th"
    )


def get_display_file_name(file_name: Path) -> str:
    return "stdin" if file_name == STDIN else file_name


class StdStream(StrEnum):
    STDIN = auto()
    STDOUT = auto()
    STDERR = auto()


STREAM_MODES = {StdStream.STDIN: "r", StdStream.STDOUT: "w", StdStream.STDERR: "w"}


@contextlib.contextmanager
def smart_open(filename=None, stream_type=StdStream.STDOUT):
    f"""
    context manager to open a file or use stdout depending on the filename, stdin is defined as '-'
    by corollary a std stream won't be closed on exit only file system streams
    :param filename: the filename to use '-' indicates a standard stream selected by stream_type
    :param stream_type: how to treat the stream choices are {', '.join(StdStream.__members__)}

    """

    mode = STREAM_MODES[stream_type]
    if filename and str(filename) != "-":
        fh = open(filename, mode)
    else:
        fh = sys.stdout

    try:
        yield fh
    finally:
        if fh not in (sys.stdout, sys.stdin, sys.stderr):
            fh.close()


class DebugStdout:
    """
    class to intercept stdout and print debugging info of where strings ae printed from

    usage:

        from contextlib import redirect_stdout
        with redirect_stdout(DebugStdout):
            ... # your code here
    """

    def __init__(self, in_old_stdout):
        self.old_stdout = in_old_stdout

    def write(self, msg):
        self.old_stdout.write(msg)
        from inspect import currentframe, getframeinfo

        frameinfo = getframeinfo(currentframe().f_back)

        print(f" <-[{frameinfo.filename}, {frameinfo.lineno}]", file=sys.stderr)


def flatten(in_list: Union[List[Any], List[Union[List[Any], Any]]]) -> List[Any]:
    """
    flatten a list of lists into a single list
    :param in_list: a putative lists of lists
    :return: a list of anything but lists
    """
    out = []
    for elem in in_list:
        if not isinstance(elem, str):
            out.extend(elem)
        else:
            out.append(elem)
    return out


def get_version() -> str:
    """
        get the current version of nef pipelines
    :return: a version string
    """

    file_path = Path(__file__)
    root_path = file_path.parent.parent
    version_path = root_path / "VERSION"

    with open(version_path) as file_h:
        version = file_h.read().strip()

    return version


class FStringTemplate:
    """class for makes an f string based template substitution is delayed
    until the instance is converted to a string
    """

    def __init__(self, template):
        self.template = template

    def __str__(self):
        variables = inspect.currentframe().f_back.f_globals.copy()
        variables.update(inspect.currentframe().f_back.f_locals)
        return self.template.format(**variables)


def unused_to_none(value: str) -> str:
    """
    If the value is a NEF UNUSED [.] replace it with None
    :param value: value which might be UNUSED
    :return: value or None if the value was UNUSED
    """
    from nef_pipelines.lib.nef_lib import (
        UNUSED,  # to avoid circular import, move to constants?
    )

    if value == UNUSED:
        value = None
    return value


def unused_to_empty_string(value):
    """
    If the value is a NEF UNUSED [.] replace it with the empty string
    :param value: value which might be UNUSED
    :return: value or "" if the value was UNUSED
    """
    from nef_pipelines.lib.nef_lib import (
        UNUSED,  # to avoid circular import , move to constants?
    )

    if value == UNUSED:
        value = ""
    return value


def _row_to_table(rows: Dict[str, str], headers=("tag", "value")) -> List[List[str]]:
    """
    convert a dictionary into a table with tabulate with two columns containing
    keys and values
    :param rows: a dictionary of key value pairs [tags and values...]
    :return: a formatted tabulate table as a string
    """
    values = [
        headers,
    ]
    for key, value in rows.items():
        values.append([key, value])
    value_string = tabulate(values, headers="firstrow")
    return value_string


def strings_to_table_terminal_sensitive(
    strings: str,
    used_width: int = 0,
    min_width: int = 20,
    fallback_terminal_width: int = 100,
):
    """
    given a list of strings convert them to a table where the number of columns os compatible with the terminal width
    :param strings: a list of strings to tabulate
    :param used_width: any width of the terminal already used
    :param min_width: the minum width for a table column
    :param fallback_terminal_width: if the terminal width can't be determined use this width
    :return: a list of list where each sub list is a row, which is  suitable for formatting with tabulate.tabulate
    """
    try:
        width, _ = os.get_terminal_size()
    except Exception:
        width = fallback_terminal_width

    width -= used_width

    # apply a sensible minimum width
    if width < min_width:
        width = min_width

    if len(strings) > 0:
        frame_name_widths = [len(frame_name) for frame_name in strings]
        max_frame_name_width = max(frame_name_widths)

        columns = int(floor(width / (max_frame_name_width + 1)))
        column_width = int(floor(width / columns))

        columns = 1 if columns == 0 else columns

        strings = [frame_name.rjust(column_width) for frame_name in strings]
        table_list = chunks(strings, columns)
    else:
        table_list = [
            [],
        ]

    return list(table_list)


def strings_to_tabulated_terminal_sensitive(
    strings: str,
    used_width: int = 0,
    min_width: int = 20,
    fallback_terminal_width: int = 100,
):

    table = strings_to_table_terminal_sensitive(
        strings, used_width, min_width, fallback_terminal_width
    )
    return tabulate(table, tablefmt="plain")


def strip_characters_left(target: str, letters: str) -> Tuple[str, str]:
    """
     strip the characters from letters from the left of the target and return a tuple containing
        1. the string without the stripped letters
        2. the letters that were stripped off

    :param target: the string to strip
    :param letters:  the letters to be stripper
    :return: a tuple of the form (<stripped-string>, <stripped-letters>)
    """

    remaining = target.lstrip(letters)

    num_stripped = len(target) - len(remaining)
    stripped = target[:num_stripped] if num_stripped > 0 else ""

    return stripped, remaining


def strip_characters_right(target: str, letters: str) -> Tuple[str, str]:
    """
    strip the characters from letters from the right of the target and return a tuple containing
        1. the string without the stripped letters
        2. the letters that were stripped off

    :param target: the string to strip
    :param letters:  the letters to be stripper
    :return: a tuple of the form (<stripped-string>, <stripped-letters>)
    """

    remaining = target.rstrip(letters)
    num_stripped = len(target) - len(remaining)
    stripped = target[-num_stripped:] if num_stripped > 0 else ""

    return remaining, stripped


def strip_line_comment(line: str, comment_character: str = "#") -> Tuple[str, str]:
    comment = None
    if comment_character in line:
        index = line.index(comment_character)
        comment = line[index + 1 :]
        line = line[:index]

    return line, comment


# https://stackoverflow.com/questions/480214/how-do-i-remove-duplicates-from-a-list-while-preserving-order
def remove_duplicates_stable(seq: Iterable) -> List:
    """
    remove  duplicates from a list while preseving its order, so for example
    [1 2 2 3 45 2 1 6] would yield 1 2 3 4 5 6
    :param seq: the iteratble to remove duplicates from
    :return: the original iterable without the duplicates

    TODO: could this be recast as an iterable returning an iterable
    """

    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


# https://stackoverflow.com/questions/21303224/iterate-over-all-pairs-of-consecutive-items-in-a-list
def iter_consecutive_pairs(seq: Iterator) -> Iterator[Tuple]:
    """
    iterate over consecutive pairs from a list, for example for the list [1, 7, 3, 5]
    this will yield the items (1,7), (7,3), (3,5)
    :param seq: the sequence to iterate pairs from
    :return: an iteration over the consecutive pairs as tuples
    """
    for first, second in zip(seq, seq[1:]):
        yield first, second


# https://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically
string_types = (type(b""), type(""))


def deprecated(reason):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter("always", DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                warnings.simplefilter("default", DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))


def convert_to_float_or_exit(putative_float: str, line_info: LineInfo) -> float:
    """
    convert input string to a float or call exit_error with a reasonable error message
    :param putative_float: the string that could be a float
    :param line_info: line info of where the string came from for error message
    :return: the string converted to a float
    """
    if not is_float(putative_float):
        msg = f"""
                the text {putative_float} could not be converted to a float, at
                line: {line_info.line_no} in
                file: {line_info.file_name} the line was
                line: {line_info.line}
            """
        exit_error(msg)
    return float(putative_float)


def fnmatch_one_of(target: str, patterns: Tuple[str, ...]) -> bool:
    """
    Check if a string code matches any of the patterns

    Args:
        target: the string to match
        patterns: a tuple of patterns to match using fnmatch

    Returns:
        True if the chain code matches any of the patterns
    """
    return any(fnmatch(target, pattern) for pattern in patterns)


def exit_if_file_has_bytes_and_no_force(output_file: Path, force: bool):
    """
    Exit if the output file exists is not empty and the force flag is not set

    Args:
        output_file: the output file
        force: if the file exists and isn't empty it still isn't an error if this flag is set
    """
    if output_file != STDOUT:
        if file_exists_and_has_bytes(output_file) and not get_global("force", force):
            msg = f"""
            the file {output_file} already exists and is not empty, if you want to overwrite it use the --force flag
            """
            exit_error(msg)


def file_exists_and_has_bytes(output_file: Path):
    """
    Check if a file exists and has bytes in it

    Args:
        output_file: the file to check

    Returns:
        True if the file exists and has bytes in it
    """
    return (
        output_file.exists()
        and output_file.is_file()
        and output_file.stat().st_size != 0
    )



# TODO: xplor names -> iupac and deal with %% and ## properly
# TODO: add common experiment types
# TODO: add a chemical shift list reference
# TODO: _nef_nmr_spectrum: value_first_point, folding, absolute_peak_positions, is_acquisition
# TODO: cleanup
# TODO: multiple assignments per peak... howto in nef
# TODO: add libs pipeline

import itertools
import string
import sys
from collections import Counter, OrderedDict
from enum import auto
from pathlib import Path
from typing import Dict, List, Tuple

import typer
from fyeah import f
from pynmrstar import Entry, Loop, Saveframe
from strenum import LowercaseStrEnum

from nef_pipelines.lib import constants
from nef_pipelines.lib.constants import NEF_UNKNOWN
from nef_pipelines.lib.isotope_lib import GAMMA_RATIOS
from nef_pipelines.lib.nef_lib import (
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import (
    MoleculeType,
    get_residue_name_from_lookup,
    sequence_from_entry,
    translate_1_to_3,
)
from nef_pipelines.lib.structures import (
    UNUSED,
    AtomLabel,
    LineInfo,
    PeakAxis,
    PeakList,
    PeakListData,
    PeakValues,
    Residue,
    SequenceResidue,
)
from nef_pipelines.lib.util import (
    NEWLINE,
    STDIN,
    exit_error,
    is_float,
    is_int,
    parse_comma_separated_options,
    strip_characters_left,
)
from nef_pipelines.transcoders.nmrview import import_app

from ..nmrview_lib import parse_float_list, parse_tcl

app = typer.Typer()


class ResidueHandlingOption(LowercaseStrEnum):

    STOP = auto()  # stop processing
    PSEUDO = (
        auto()
    )  # continue processing but the residue becomes a pseudo residue with sequence code @xxx
    UNASSSIGN = (
        auto()
    )  # set the chain_code, sequence code and  residue type to . and continue
    CONTINUE = auto()  # don't change anything take whats in the file

    CASE = (
        auto()
    )  # if the residue name matches the sequence but the case is different correct it
    WARN = auto()  # report the peak with missing residue information


RESIDUE_HANDLING_HELP = """
what to do when the residue for a peak isn't defined in the sequence (if read) but it has a chain and residue number and
the residue name doesn't match the one in the sequence. Choices are
stop: stop processing if the residue name isn't defined or is mismatched
pseudo: continue processing but the residue becomes a pseudo residue with sequence_code @xxx and the residue name read
unassign: set the chain_code, sequence code and  residue type to . and continue without warning
continue: leave everything it as it is

two modifiers can be added to the choice by adding them as extra options

warn: report the peak with mismatched residue information on STDERR, use the read values and continue processing
case: if the residue name matches the sequence but the case is different correct it and continue
"""


class ResidueNameTypeOption(LowercaseStrEnum):
    AUTO = auto()
    SINGLE = auto()
    THREE = auto()


RESIDUE_NAME_TYPE_HELP = """
auto: translate name based on its length otherwise, single: single letter amino acid code
three: three letter or longer amino acid code
"""


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    in_file: Path = typer.Option(
        STDIN, "-i", "--in", help="file to read nef data from", metavar="<NEF-FILE>"
    ),
    chain_codes: List[str] = typer.Option(
        None, "--chain", help="chain code", metavar="<chain-code>"
    ),
    residue_handling: List[ResidueHandlingOption] = typer.Option(
        None, "--residue-handling", help=RESIDUE_HANDLING_HELP
    ),
    residue_name_type: ResidueNameTypeOption = typer.Option(
        ResidueNameTypeOption.AUTO, "--residue-name-type", help=RESIDUE_NAME_TYPE_HELP
    ),
    molecule_type: MoleculeType = typer.Option(
        MoleculeType.PROTEIN,
        "--molecule-type",
        help="molecule type one of protein, dna, rna or  carbohydrate",
    ),
    entry_name: str = typer.Option(
        "nmrview", "-n", "--name", help="entry name", metavar="<entry-name>"
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input peak files", metavar="<peak-file.xpk>"
    ),
):
    """convert nmrview peak file <nmrview>.xpk files to NEF [alpha for new residue handling]"""

    if not chain_codes:
        chain_codes = ["A"] * len(file_names)

    if len(chain_codes) == 1:
        chain_codes = chain_codes * len(file_names)

    chain_codes = parse_comma_separated_options(chain_codes)

    entry = read_or_create_entry_exit_error_on_bad_file(in_file, entry_name)

    sequence = sequence_from_entry(entry)

    _exit_if_num_chains_and_files_dont_match(chain_codes, file_names)

    residue_handling = parse_comma_separated_options(residue_handling)

    residue_handling = _fixup_residue_handling_or_exit_bad(residue_handling)

    entry = pipe(
        entry,
        file_names,
        chain_codes,
        sequence,
        molecule_type,
        residue_name_type,
        residue_handling,
    )

    print(entry)


def pipe(
    entry: Entry,
    file_names: List[str],
    chain_codes: List[str],
    sequence: List[Residue],
    molecule_type: MoleculeType,
    residue_name_type: ResidueNameTypeOption,
    residue_handling: ResidueHandlingOption,
) -> Entry:

    frames = []
    sequence_lookup = _sequence_to_residue_type_lookup(sequence)

    frame_names_and_peak_lists = []
    for file_name, chain_code in zip(file_names, chain_codes):
        with open(file_name, "r") as lines:
            peaks_list = read_raw_peaks(
                lines,
                chain_code,
                sequence_lookup,
                file_name,
                molecule_type,
                residue_name_type,
                residue_handling,
            )

        frame_name = _make_peak_list_frame_name(peaks_list)

        frame_names_and_peak_lists.append((frame_name, peaks_list))

    frame_names_and_peak_lists = _disambiguate_frame_names(frame_names_and_peak_lists)

    for (frame_name, peaks_list), chain_code in zip(
        frame_names_and_peak_lists, chain_codes
    ):
        frames.append(_create_spectrum_frame(frame_name, peaks_list, chain_code))

    return add_frames_to_entry(entry, frames)


def _disambiguate_frame_names(peak_lists_and_entry_names):
    seen_frames = Counter()
    for i, (frame_name, peaks_list) in enumerate(peak_lists_and_entry_names):
        if frame_name in seen_frames:
            seen_frames[frame_name] += 1
            frame_name = f"{frame_name}`{seen_frames[frame_name]}`"
            peak_lists_and_entry_names[i] = (frame_name, peaks_list)
        else:
            seen_frames[frame_name] = 1
    return peak_lists_and_entry_names


def _create_entry_names_from_template_if_required(
    entry_name_template, entry_names, file_names
):
    new_entry_names = []
    for file_name, entry_name in itertools.zip_longest(file_names, entry_names):
        if entry_name:
            new_entry_names.append(entry_name)
        else:
            file_name = str(Path(file_name).stem)
            numbers_and_letters = string.ascii_lowercase + string.digits
            file_name = "".join(
                [
                    letter if letter in numbers_and_letters else "_"
                    for letter in file_name
                ]
            )
            file_name = f"_{file_name}"
            new_entry_names.append(f(entry_name_template))
    return new_entry_names


def read_raw_peaks(
    lines,
    chain_code,
    sequence_lookup,
    file_name,
    molecule_type,
    residue_name_type,
    residue_handling,
):

    header = get_header_or_exit(lines)

    header_data = read_header_data(lines, header)

    column_indices = read_peak_columns(lines, header_data)

    raw_peaks = _read_peak_data(
        lines,
        header_data,
        column_indices,
        chain_code,
        sequence_lookup,
        file_name,
        molecule_type,
        residue_name_type,
        residue_handling,
    )

    return PeakList(header_data, raw_peaks)


def _read_peak_data(
    lines,
    header_data,
    column_indices,
    chain_code,
    sequence_lookup,
    file_name,
    molecule_type,
    residue_name_type,
    residue_handling,
):
    raw_peaks = []
    field = None
    axis_index = None
    for line_no, raw_line in enumerate(lines):
        line_info = LineInfo(file_name, line_no, raw_line)
        if not len(raw_line.strip()):
            continue
        try:
            peak = {}
            tcl_list = parse_tcl(raw_line)
            # TODO validate and report errors
            peak_index = int(tcl_list[0])

            for axis_index, axis in enumerate(header_data.axis_labels):
                axis_values = _read_axis_for_peak(
                    tcl_list,
                    axis,
                    column_indices,
                    chain_code,
                    sequence_lookup,
                    molecule_type,
                    residue_name_type,
                    residue_handling,
                    line_info,
                )

                peak[axis_index] = PeakAxis(*axis_values)

            raw_values = _read_values_for_peak(tcl_list, column_indices)
            peak["values"] = PeakValues(peak_index, **raw_values)

            raw_peaks.append(peak)

        except Exception as e:
            field = str(field) if field else "unknown"
            msg = (
                f"failed to parse the line {line_no} with input: '{raw_line.strip()}' field: {field}  axis: "
                f"  {axis_index + 1} exception: {e}"
            )
            exit_error(msg)
    return raw_peaks


def _exit_if_num_chains_and_files_dont_match(chain_codes, file_names):
    if len(chain_codes) != len(file_names):
        msg = f"""
            the number of chain codes {len(chain_codes)} does not match number of files {len(file_names)}
            the chain codes are {', '.join(chain_codes)}
            the files are
            {NEWLINE.join(file_names)}
        """
        exit_error(msg)


def read_peak_columns(lines, header_data):
    line = next(lines)
    raw_headings = line.split()
    heading_indices = OrderedDict({"index": 0})
    for axis_index, axis in enumerate(header_data.axis_labels):
        for axis_field in list("LPWBEJU"):
            header = f"{axis}.{axis_field}"
            if header in raw_headings:
                heading_indices[header] = raw_headings.index(header) + 1
    for peak_item in ["vol", "int", "stat", "comment", "flag0"]:
        if peak_item in raw_headings:
            heading_indices[peak_item] = raw_headings.index(peak_item) + 1
    return heading_indices


def read_header_data(lines, headers):
    data_set = None
    sweep_widths = []
    spectrometer_frequencies = []
    num_axis = None
    axis_labels = None
    for header_no, header_type in enumerate(headers):
        line = next(lines)
        if header_type == "label":
            axis_labels = line.strip().split()
            num_axis = len(axis_labels)
        elif header_type == "dataset":
            data_set = line.strip()
        elif header_type == "sw":
            line_no = header_no + 2
            sweep_widths = parse_float_list(line, line_no)
            check_num_fields(sweep_widths, num_axis, "sweep widths", line, line_no)
        elif header_type == "sf":
            line_no = header_no + 2
            spectrometer_frequencies = parse_float_list(line, line_no)
            check_num_fields(
                spectrometer_frequencies,
                num_axis,
                "spectrometer frequencies",
                line,
                line_no,
            )

    # sweep widths are in ppm for nef!
    sweep_widths = [float(sweep_width) for sweep_width in sweep_widths]
    spectrometer_frequencies = [
        float(spectrometer_frequency)
        for spectrometer_frequency in spectrometer_frequencies
    ]

    for i, (sweep_width, spectrometer_frequency) in enumerate(
        zip(sweep_widths, spectrometer_frequencies)
    ):
        sweep_widths[i] = f"{sweep_width / spectrometer_frequency:.4f}"

    # TODO: peaks shifts, spectrometer frequencies how many decimal points
    spectrometer_frequencies = [
        f"{spectrometer_frequency:4}"
        for spectrometer_frequency in spectrometer_frequencies
    ]

    peak_list_data = PeakListData(
        num_axis, axis_labels, data_set, sweep_widths, spectrometer_frequencies
    )
    return peak_list_data


def get_header_or_exit(lines):
    header_items = ["label", "dataset", "sw", "sf"]

    line = next(lines)

    headers = []
    if line:
        headers = line.strip().split()

    if len(headers) != 4:
        msg = f"""this doesn't look like an nmrview xpk file,
                  i expected a header containing 4 items on the first line: {','.join(header_items)}
                  i got {line} at line 1"""
        exit_error(msg)

    for name in header_items:
        if name not in headers:
            msg = f"""this doesn't look like an nmrview xpk file,
                       i expected a header containing the values: {', '.join(header_items)}
                       i got '{line}' at line 1"""
            exit_error(msg)

    return headers


def _exit_if_nmrview_residue_name_doesnt_match_sequence(
    nmrview_residue_name, chain_code, residue_number, residue_type, line_info
):

    if (
        residue_type != NEF_UNKNOWN
        and nmrview_residue_name
        and nmrview_residue_name != residue_type
    ):
        msg = f"""
                in file {line_info.file_name} at line {line_info.line_no}
                for residue number {residue_number} in chain {chain_code} the residue type in the NEF sequence
                [{residue_type}] does not match the one in the nmrview file [{nmrview_residue_name}]
                the line was
                 {line_info.line}

                to ignore this error run with --ignore-bad-residues which which will use residue names from the NEF
                sequence
            """
        exit_error(msg)


def _read_axis_for_peak(
    tcl_list,
    axis,
    heading_indices,
    chain_code,
    sequence_lookup,
    molecule_type,
    residue_name_type,
    residue_handling,
    line_info,
):
    # print(line_info.line, ':', axis, heading_indices)
    axis_values = []
    for axis_field in list("LPWBEJU"):
        header = f"{axis}.{axis_field}"
        field_index = heading_indices[header]
        value = tcl_list[field_index]
        if axis_field == "L":
            axis_values.append(
                _read_atom_label(
                    value,
                    chain_code,
                    sequence_lookup,
                    molecule_type,
                    residue_name_type,
                    residue_handling,
                    line_info,
                )
            )
        elif axis_field == "P":
            axis_values.append(_read_shift(value, line_info))
        elif axis_field in "WJU":
            pass
        elif axis_field == "E":
            merit = value
            axis_values.append(merit)
    return axis_values


def _read_shift(value, line_info):
    if not is_float(value):
        msg = f"""
                    in file {line_info.file_name} at the line {line_info.line_no}
                    i expected a float for the peak position, but i got the value '{value}'
                    the line was
                    {line_info.line}
                """
        exit_error(msg)
    shift = float(value)
    return shift


def _read_atom_label(
    value,
    chain_code,
    sequence_lookup,
    molecule_type,
    residue_name_type,
    residue_handling,
    line_info,
):
    label = value[0] if value else "?"
    label = label.strip()

    if "." in label and len(label.split(".")) == 2:
        residue_name_residue_number, atom_name = label.split(".")
        file_residue_name, sequence_code = strip_characters_left(
            residue_name_residue_number, string.ascii_letters
        )
    else:
        sequence_code = None
        atom_name = None
        file_residue_name = None if label == "?" else label

    _exit_if_sequence_code_isnt_int_or_none(line_info, sequence_code)

    if sequence_code:
        if is_int(sequence_code):
            sequence_code = int(sequence_code)

    if sequence_code:
        sequence_residue_name = get_residue_name_from_lookup(
            chain_code, sequence_code, sequence_lookup
        )
    else:
        sequence_residue_name = None

    if file_residue_name and sequence_code:
        file_residue_name = _translate_1_to_3_or_or_exit_bad(
            file_residue_name, molecule_type, residue_name_type, line_info
        )

    file_residue_name = _correct_residue_name_case_if_required(
        residue_handling, file_residue_name, sequence_residue_name
    )

    _warn_of_residue_name_mismatch_if_required(
        file_residue_name, sequence_residue_name, residue_handling, line_info
    )

    residue_handling = [
        option
        for option in residue_handling
        if option not in {ResidueHandlingOption.CASE, ResidueHandlingOption.WARN}
    ]

    residue_handling = residue_handling.pop()

    if (
        file_residue_name
        and sequence_residue_name
        and (sequence_residue_name != file_residue_name)
    ):
        if residue_handling == ResidueHandlingOption.STOP:
            _exit_residue_name_mismatch(
                file_residue_name, sequence_residue_name, line_info
            )
            residue_name = sequence_residue_name
        elif residue_handling == ResidueHandlingOption.PSEUDO:
            residue_name = file_residue_name
            sequence_code = f"@{sequence_code}"
        elif residue_handling == ResidueHandlingOption.UNASSIGN:
            residue_name = None
            sequence_code = None
            chain_code = None
            atom_name = None
        elif residue_handling == ResidueHandlingOption.SEQUENCE:
            residue_name = sequence_residue_name
        elif residue_handling == ResidueHandlingOption.CONTINUE:
            residue_name = file_residue_name
        else:
            raise Exception("unexpected")
    elif not file_residue_name and sequence_residue_name:
        residue_name = sequence_residue_name
    else:
        residue_name = file_residue_name

    chain_code = chain_code if sequence_code or atom_name else UNUSED
    sequence_code = UNUSED if not sequence_code else sequence_code
    atom_name = UNUSED if not atom_name else atom_name
    residue_name = UNUSED if not residue_name else residue_name

    atom = AtomLabel(
        SequenceResidue(chain_code, sequence_code, residue_name),
        atom_name,
    )

    return atom


def _warn_of_residue_name_mismatch_if_required(
    file_residue_name, sequence_residue_name, residue_handling, line_info
):
    if ResidueHandlingOption.WARN in residue_handling:
        msg = (
            f"WARNING: at line {line_info.line_no} in file {line_info.file_name} the residue name "
            + f"{file_residue_name} doesn't match the sequence residue name {sequence_residue_name}"
        )
        print(msg, file=sys.stderr)


def _correct_residue_name_case_if_required(
    residue_handling, file_residue_name, sequence_residue_name
):
    if ResidueHandlingOption.CASE in residue_handling:
        if (
            sequence_residue_name
            and file_residue_name
            and sequence_residue_name.lower() == file_residue_name.lower()
        ):
            file_residue_name = sequence_residue_name
    return file_residue_name


def _exit_residue_name_mismatch(file_residue_name, sequence_residue_name, line_info):
    msg = f"""
                    in file {line_info.file_name} at line {line_info.line_no}
                    the residue name {file_residue_name} doesn't match the one in the sequence {sequence_residue_name}
                    the line was
                    {line_info.line}
                """
    exit_error(msg)


def _exit_no_residue_name(line_info):
    msg = f"""
            in file {line_info.file_name} at line {line_info.line_no}
            residue name is not defined for the peak
            the line was
            {line_info.line}
        """
    exit_error(msg)


def _translate_1_to_3_or_or_exit_bad(
    residue_name, molecule_type, residue_name_type, line_info
):
    if residue_name:
        if residue_name_type == ResidueNameTypeOption.AUTO:
            if len(residue_name) > 1:
                residue_name = translate_1_to_3(residue_name, molecule_type)
        elif residue_name_type == ResidueNameTypeOption.SINGLE:
            if len(residue_name) > 1:
                msg = f"""
                    in file {line_info.file_name} at line {line_info.line_no}
                    the residue name {residue_name} is longer than one character
                    and you have chosen to use single letter residue names
                    the line was
                    {line_info.line}
                """
                exit_error(msg)
            residue_name = translate_1_to_3(residue_name, molecule_type)

    return residue_name


def _exit_if_sequence_code_isnt_int_or_none(line_info, residue_number):
    if residue_number and not is_int(residue_number):
        msg = f"""
            in file {line_info.file_name} at line {line_info.line_no}
            I expected an integer for the residue number nut got {residue_number}
            in line
            {line_info.line}
        """
        exit_error(msg)


def _exit_if_sequence_code_not_in_the_sequence(
    chain_code, residue_number, chain_start, chain_end, line_info
):
    if residue_number and (residue_number < chain_start or residue_number > chain_end):
        msg = f"""
            in the file {line_info.file_name} at {line_info.line_no}
            a residue type is not defined for chain: {chain_code} and residue number: {residue_number}
            the line was
            {line_info.line}
        """
        exit_error(msg)


def _read_values_for_peak(line, heading_indices):
    peak_values = {}
    for value_field in ["vol", "int", "stat", "comment", "flag0"]:
        field_index = heading_indices[value_field]
        value = line[field_index]

        if value_field == "vol":
            peak_values["volume"] = float(value)
        elif value_field == "int":
            peak_values["height"] = float(value)
        elif value_field == "stat":
            peak_values["deleted"] = int(value) < 0
        elif value_field == "comment":
            comment = value[0].strip("'") if value else ""
            peak_values["comment"] = comment
        elif value_field == "flag0":
            pass

    return peak_values


def check_num_fields(fields, number, field_type, line, line_no):
    if len(fields) != number:
        msg = f"Expected {number} {field_type} got {len(fields)} for line: {line} at line {line_no}"
        exit_error(msg)


def _sequence_to_residue_type_lookup(
    sequence: List[SequenceResidue],
) -> Dict[Tuple[str, int], str]:
    result: Dict[Tuple[str, int], str] = {}
    for residue in sequence:
        result[residue.chain_code, residue.sequence_code] = residue.residue_name
    return result


def _get_isotope_code_or_exit(axis, axis_codes):
    if axis >= len(axis_codes):
        msg = f"can't find isotope code for axis {axis + 1} got axis codes {','.join(axis_codes)}"
        exit_error(msg)
    axis_code = axis_codes[axis]
    return axis_code


def round_to_nearest_and_distance(number, nearest):
    rounded = round(number / nearest) * nearest
    return rounded, abs(rounded - number)


def _guess_spectrometer_frequency(peak_list):
    frequencies = [
        float(frequency)
        for frequency in peak_list.peak_list_data.spectrometer_frequencies
    ]
    max_frequency = max(frequencies)

    divisor = 10 if max_frequency < 240 else 50

    distance_and_frequency = {}
    for spectrometer_frequency in frequencies:
        rounded_spectrometer_frequency, distance = round_to_nearest_and_distance(
            spectrometer_frequency, divisor
        )
        distance_and_frequency[distance] = spectrometer_frequency

    min_distance = min(distance_and_frequency.keys())
    return distance_and_frequency[min_distance]


def _spectrometer_frequencies_to_axis_codes(spectrometer_frequency, peak_list):
    isotopes = []
    for frequency in peak_list.peak_list_data.spectrometer_frequencies:
        frequency = float(frequency)
        ratio = frequency / spectrometer_frequency
        ratio_distances = {
            isotope: abs(ratio - gamma_ratio)
            for isotope, gamma_ratio in GAMMA_RATIOS.items()
        }
        closest_isotope = min(ratio_distances, key=ratio_distances.get)
        isotopes.append(closest_isotope)

    return isotopes


def _create_spectrum_frame(entry_name, peak_list, chain_code):

    spectrometer_frequency = _guess_spectrometer_frequency(peak_list)
    axis_isotopes = _spectrometer_frequencies_to_axis_codes(
        spectrometer_frequency, peak_list
    )

    category = "nef_nmr_spectrum"
    frame_code = f"{category}_{entry_name}"
    frame = Saveframe.from_scratch(frame_code, category)

    frame.add_tag("sf_category", category)
    frame.add_tag("sf_framecode", frame_code)
    frame.add_tag("num_dimensions", peak_list.peak_list_data.num_axis)
    frame.add_tag("chemical_shift_list", constants.NEF_UNKNOWN)
    loop = Loop.from_scratch("nef_spectrum_dimension")
    frame.add_loop(loop)
    list_tags = (
        "dimension_id",
        "axis_unit",
        "axis_code",
        "spectrometer_frequency",
        "spectral_width",
        "value_first_point",
        "folding",
        "absolute_peak_positions",
        "is_acquisition",
    )
    loop.add_tag(list_tags)
    list_data = peak_list.peak_list_data
    for i in range(list_data.num_axis):
        row = {
            "dimension_id": i + 1,
            "axis_unit": "ppm",
            "axis_code": axis_isotopes[i],
            "spectrometer_frequency": list_data.spectrometer_frequencies[i],
            "spectral_width": (
                list_data.sweep_widths[i] if list_data.sweep_widths else NEF_UNKNOWN
            ),
            "value_first_point": NEF_UNKNOWN,
            "folding": "circular",
            "absolute_peak_positions": "true",
            "is_acquisition": NEF_UNKNOWN,
        }
        loop.add_data(
            [
                row,
            ]
        )

    loop = Loop.from_scratch("nef_spectrum_dimension_transfer")
    frame.add_loop(loop)
    transfer_dim_tags = ("dimension_1", "dimension_2", "transfer_type")
    loop.add_tag(transfer_dim_tags)
    loop = Loop.from_scratch("nef_peak")
    frame.add_loop(loop)

    # TODO: put this in a sane order!
    peak_tags = [
        "index",
        "peak_id",
        "volume",
        "volume_uncertainty",
        "height",
        "height_uncertainty",
    ]
    position_tags = [
        (f"position_{i + 1}", f"position_uncertainty_{i + 1}")
        for i in range(list_data.num_axis)
    ]
    position_tags = itertools.chain(*position_tags)
    atom_name_tags = [
        (
            f"chain_code_{i + 1}",
            f"sequence_code_{i + 1}",
            f"residue_name_{i + 1}",
            f"atom_name_{i + 1}",
        )
        for i in range(list_data.num_axis)
    ]
    atom_name_tags = itertools.chain(*atom_name_tags)
    tags = [*peak_tags, *position_tags, *atom_name_tags]

    loop.add_tag(tags)
    for i, peak in enumerate(peak_list.peaks):
        peak_values = peak["values"]
        if peak_values.deleted:
            continue

        positions = {}
        for tag in tags:
            if tag.split("_")[0] == "position" and len(tag.split("_")) == 2:
                index = int(tag.split("_")[-1]) - 1
                positions[tag] = peak[index].ppm

        chain_codes = {}
        for tag in tags:
            if tag.split("_")[:2] == ["chain", "code"]:
                index = int(tag.split("_")[-1]) - 1
                chain_code = peak[index].atom_labels.residue.chain_code
                chain_code = chain_code if chain_code is not None else chain_code
                chain_code = chain_code if chain_code else "."
                chain_codes[tag] = chain_code

        sequence_codes = {}
        for tag in tags:
            if tag.split("_")[:2] == ["sequence", "code"]:
                index = int(tag.split("_")[-1]) - 1
                sequence_code = peak[index].atom_labels.residue.sequence_code
                sequence_code = sequence_code if sequence_code else "."
                sequence_codes[tag] = sequence_code

        residue_names = {}
        for tag in tags:
            if tag.split("_")[:2] == ["residue", "name"]:
                index = int(tag.split("_")[-1]) - 1

                # TODO: there could be more than 1 atom label here and this should be a list...
                residue_name = peak[index].atom_labels.residue.residue_name
                residue_name = residue_name if residue_name else "."
                residue_names[tag] = residue_name

        atom_names = {}
        for tag in tags:
            if tag.split("_")[:2] == ["atom", "name"]:
                index = int(tag.split("_")[-1]) - 1

                atom_name = peak[index].atom_labels.atom_name
                atom_name = atom_name if atom_name else "."
                atom_names[tag] = atom_name

        row_dict = {
            "index": i + 1,
            "peak_id": peak_values.serial,
            **chain_codes,
            **sequence_codes,
            **residue_names,
            **atom_names,
            **positions,
            "volume": peak_values.volume,
            "volume_uncertainty": NEF_UNKNOWN,
            "height": peak_values.height,
            "height_uncertainty": NEF_UNKNOWN,
        }
        loop.add_data(
            [
                row_dict,
            ]
        )
    return frame


# TODO: can be replaced by str.removesuffix when min python version >= 3.9
def _remove_suffix(string: str, suffix: str) -> str:

    result = string
    if string.endswith(suffix):
        result = string[: -len(suffix)]

    return result


def _make_peak_list_frame_name(peaks_list):
    entry_name = peaks_list.peak_list_data.data_set.replace(" ", "_")
    entry_name = _remove_suffix(entry_name, ".nv")
    entry_name = entry_name.replace(".", "_")
    return entry_name


def _fixup_residue_handling_or_exit_bad(residue_handling):

    number_residue_handling_options = len(residue_handling)
    if number_residue_handling_options == 0:
        residue_handling = [
            ResidueHandlingOption.STOP,
        ]

    active_residue_options = _remove_residue_handling_modifiers(residue_handling)
    if len(active_residue_options) != 1:
        _exit_too_many_residue_handling_options(residue_handling)
    return residue_handling


def _remove_residue_handling_modifiers(residue_handling):
    return [
        option
        for option in residue_handling
        if option not in {ResidueHandlingOption.CASE, ResidueHandlingOption.WARN}
    ]


def _exit_too_many_residue_handling_options(residue_handling):
    active_residue_handling_options = _remove_residue_handling_modifiers(
        residue_handling
    )
    number_active_residue_handling_options = len(active_residue_handling_options)
    if number_active_residue_handling_options > 1:
        msg = f"""
            you have specified residue handling options  {','.join(active_residue_handling_options)}
            that are not compatible with each other, you can only combine options with case and warn...
            """
    exit_error(msg)



from pathlib import Path
from typing import Dict, Iterable, List, Tuple

import typer

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import (
    BadResidue,
    get_chain_code_iter,
    get_residue_name_from_lookup,
    sequence_from_entry_or_exit,
    sequence_to_residue_name_lookup,
    translate_1_to_3,
)
from nef_pipelines.lib.shift_lib import shifts_to_nef_frame
from nef_pipelines.lib.structures import (
    AtomLabel,
    LineInfo,
    SequenceResidue,
    ShiftData,
    ShiftList,
)
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    is_int,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.sparky import import_app

SHIFT = "Shift"
ATOM = "Atom"
GROUP = "Group"
NUCLEUS = "Nuc"

EXPECTED_HEADINGS = set(f"{GROUP} {ATOM} {NUCLEUS} {SHIFT} SDev Assignments".split())


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def shifts(
    chain_codes: List[str] = typer.Option(
        None,
        "--chains",
        help="chain codes as a list of names separated by commas, repeated calls will add further chains [default A]",
        metavar="<CHAIN-CODES>",
    ),
    frame_name: str = typer.Option(
        "sparky", "-f", "--frame-name", help="a name for the frame"
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type shifts.txt", metavar="<SPARKY-shifts>.txt"
    ),
):
    """- convert sparky shift file <sparky-shifts>.txt to NEF"""

    chain_codes = parse_comma_separated_options(chain_codes)

    if not chain_codes:
        chain_codes = ["A"]
    elif len(chain_codes) == 1 and len(file_names) > 1:
        chain_codes = chain_codes * len(file_names)
    else:
        _exit_if_number_chain_codes_and_file_names_dont_match(chain_codes, file_names)

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    entry = pipe(entry, chain_codes, frame_name, file_names)

    print(entry)


def _exit_if_number_chain_codes_and_file_names_dont_match(chain_codes, file_names):
    if len(chain_codes) != len(file_names):
        msg = f"""\
            the number of chains [{len(chain_codes)}] doesn't match the number of files [{len(file_names)}]
            chains were {chain_codes}
            files were {file_names}
        """
        exit_error(msg)


def pipe(entry, chain_codes, frame_name, file_names):

    sequence = sequence_from_entry_or_exit(entry)

    sparky_frames = []

    chain_code_iter = get_chain_code_iter(chain_codes)
    for file_name, chain_code in zip(file_names, chain_code_iter):

        with open(file_name) as lines:

            chain_seqid_to_type = sequence_to_residue_name_lookup(sequence)

            sparky_shifts = _parse_shifts(
                lines, chain_seqid_to_type, chain_code=chain_code, file_name=file_name
            )

            frame = shifts_to_nef_frame(sparky_shifts, frame_name)

            sparky_frames.append(frame)

    return add_frames_to_entry(entry, sparky_frames)


def _convert_residue_type_to_3_let_or_exit(residue_type, line_info):
    try:
        residue_type = translate_1_to_3([residue_type])[0]
    except BadResidue:
        msg = f"""\
            unknown residue name {residue_type}
            at line {line_info.line_no} in
            file {line_info.file_name}, the line had the following content
            {line_info.line}
        """
        exit_error(msg)

    return residue_type


def _exit_if_residue_doesnt_equal_sequence(residue, chain_seqid_to_type, line_info):

    chain_code = residue.chain_code
    sequence_code = residue.sequence_code
    sequence_residue_name = get_residue_name_from_lookup(
        chain_code, sequence_code, chain_seqid_to_type
    )
    if sequence_residue_name.lower() != residue.residue_name.lower():
        msg = f"""\
            for residue number {sequence_code} in chain {chain_code} the residue type read from file
            {residue.residue_name} doesn't match the one from the sequence {sequence_residue_name}
            while reading line {line_info.line_no} from file:
            {line_info.file_name}
            the value of the line was
            {line_info.line}
        """
        exit_error(msg)


def _parse_shifts(
    lines: Iterable[str],
    chain_seqid_to_type: Dict[Tuple[str, int], str],
    chain_code: str = "A",
    file_name="unknown",
) -> ShiftList:

    heading_indices = {}

    read_shifts = []
    for i, line in enumerate(lines):
        line_info = LineInfo(file_name, i + 1, line)
        line = line.strip()

        if i == 0:
            headings = line.split()

            _exit_if_bad_header(headings, EXPECTED_HEADINGS, line_info)

            heading_indices = {
                heading: column for column, heading in enumerate(headings)
            }

        elif not line:
            continue
        else:
            shift = _parse_shifts_line(
                line_info, heading_indices, chain_code, chain_seqid_to_type
            )

            read_shifts.append(shift)

    return ShiftList(read_shifts)


def _parse_shifts_line(line_info, heading_indices, chain_code, chain_seqid_to_type):
    fields = line_info.line.split()

    _exit_if_wrong_number_data_columns(fields, heading_indices, line_info)

    fields = {heading: fields[column] for heading, column in heading_indices.items()}

    group = fields[GROUP]
    atom = fields[ATOM]
    shift = float(fields[SHIFT])
    nucleus = fields[NUCLEUS]

    residue_type = _read_residue_name_from_group_or_exit(group, line_info)

    sequence_code = read_residue_number_from_group_or_exit(
        group, residue_type, line_info
    )

    residue_type = _convert_residue_type_to_3_let_or_exit(residue_type, line_info)

    residue = SequenceResidue(chain_code, sequence_code, residue_type)

    _exit_if_residue_doesnt_equal_sequence(residue, chain_seqid_to_type, line_info)

    element, isotope_number = _parse_element_and_isotope_number(nucleus)

    atom_label = AtomLabel(
        residue, atom, element=element, isotope_number=isotope_number
    )

    shift = ShiftData(atom_label, shift)
    return shift


def _exit_if_wrong_number_data_columns(fields, headings, line_info):
    num_fields = len(fields)
    num_headings = len(headings)
    if num_fields < num_headings:
        msg = f"""\
                    at line {line_info.line_no} in
                    file {line_info.file_name}
                    there weren't enough columns of data i expected {num_headings} but got {num_fields}
                    so the number of fields matched the length of the headings.
                    the headings were {headings}
                    the line was {line_info.line}
                """
        exit_error(msg)


def _exit_if_bad_header(headings, EXPECTED_HEADINGS, line_info):
    if not set(headings).issubset(EXPECTED_HEADINGS):
        msg = f"""\
                    the file {line_info.file_name} doesn't look like a sparky file"
                    line 1 should contain the the standard sparky headings:
                    {', '.join(EXPECTED_HEADINGS)}
                    i got:
                    {line_info.line}
                """
        exit_error(msg)


def read_residue_number_from_group_or_exit(group, residue_type, line_info):

    residue_number = group[len(residue_type) :]
    if not is_int(residue_number):
        msg = f"""\
            bad sequence code  {residue_number}  [should be an integer for a spark residue number]
            at line {line_info.line_no} in
            file {line_info.file_name}, the line had the following content
            {line_info.line}
        """
        exit_error(msg)

    return int(residue_number)


def _read_residue_name_from_group_or_exit(group: str, line_info: LineInfo):

    residue_type = _get_starting_alpha(group)

    if not residue_type:
        msg = f"""\
            there was no residue type for the residues specificed on line {line_info.line_no} of
            file {line_info.file_name} the line was
            {line_info.line}
        """
        exit_error(msg)

    return residue_type


def _get_starting_alpha(group):
    result = []
    for char in group:
        if char.isalpha():
            result.append(char)
        else:
            break
    return "".join(result)


def get_starting_number(group):
    result = []
    for char in group:
        if char.isdigit():
            result.append(char)
        else:
            break
    return "".join(result)


def _parse_element_and_isotope_number(nucleus):
    isotope_number = get_starting_number(nucleus)
    if not isotope_number:
        isotope_number = UNUSED

    if not isotope_number == UNUSED:
        element = nucleus[len(isotope_number) :]

    if not element:
        isotope_number = UNUSED

    return element, isotope_number



from pathlib import Path
from sys import stdout
from typing import Dict, List

import typer
from fyeah import f

from nef_pipelines.lib.nef_lib import (
    is_save_frame_name_in_entry,
    molecular_system_from_entry,
    molecular_system_from_entry_or_exit,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import (
    chains_from_frames,
    get_chain_starts,
    sequences_from_frames,
)
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.util import STDIN, exit_error, in_pytest, smart_open
from nef_pipelines.transcoders.fasta.exporters.sequence import STDOUT
from nef_pipelines.transcoders.nmrview import export_app

app = typer.Typer()

OUTPUT_HELP = """\
Where to write the output to. The default is a set of files with each file is named by a template of the form
"{chain_code}.seq" where {chain_code} is replaced by the chain name. If set to a string containing {chain_code} the
string will be used as a template for the filenames If set to - output will be written to stdout and individual chains
will be written as virtual files with a banner of the form ------------- chain_code.seq ------------- separating them.
A template starting with - will still output to stdout but will use the rest of the string as a template for the
filename
"""


@export_app.command()
def sequences(
    input: Path = typer.Option(
        STDIN, "-i", "--in", help="sequence_text to read input from [- is stdin]"
    ),
    output_destination: str = typer.Option(
        "{chain_code}.seq", "-o", "--out", help=OUTPUT_HELP
    ),
    chain_selectors: List[str] = typer.Argument(
        None,
        help="the names of the chains to export, can be a comma separated"
        " list of chains. Repeated calls add more chains. The default"
        " is to write all chains",
        metavar="<chains>",
    ),
):

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    _exit_if_no_molecular_system_in_entry(entry)

    molecular_system_frame = entry.get_saveframes_by_category("nef_molecular_system")[0]

    chain_selectors = (
        chain_selectors
        if chain_selectors
        else chains_from_frames(molecular_system_frame)
    )

    molecular_system_frame = molecular_system_from_entry_or_exit(entry)

    chain_codes = chains_from_frames(molecular_system_frame)

    _exit_if_chain_selector_not_in_chain_codes(chain_codes, chain_selectors)

    if output_destination == str(STDOUT):
        output_to_stdout = True
        file_name_template = "{chain_code}.seq"
    elif output_destination.startswith("-"):
        output_to_stdout = True
        file_name_template = output_destination[1:]
    else:
        output_to_stdout = False
        file_name_template = output_destination

    if "{chain_code}" not in file_name_template and file_name_template != str(STDOUT):
        msg = f"the file name template {file_name_template} does not contain the string {{chain_code}}"
        exit_error(msg)

    pipe(entry, output_to_stdout, file_name_template, chain_selectors)


def _exit_if_no_molecular_system_in_entry(entry):
    if not is_save_frame_name_in_entry(entry, "nef_molecular_system"):
        msg = f"the entry {entry.entry_id} does not contain a nef_molecular_system saveframe"
        exit_error(msg)


def pipe(entry, output_to_stdout, file_name_template, chain_selectors):

    chains_to_filenames = _build_chains_to_filenames(
        chain_selectors, file_name_template
    )

    chain_sequences = _nmrview_sequences_from_entry(entry, chain_selectors)

    file_banners, final_banner = _make_file_name_banners(chains_to_filenames)

    for chain_code, sequence_text in chain_sequences.items():

        output_file_name = (
            str(STDOUT) if output_to_stdout else chains_to_filenames[chain_code]
        )

        with smart_open(output_file_name) as file_pointer:

            if output_to_stdout and len(file_banners) > 1:
                print(file_banners[chain_code], file=file_pointer)

            print(sequence_text, file=file_pointer)

    if output_to_stdout and len(file_banners) > 1:
        print(final_banner, file=file_pointer)

    stdout_is_atty = stdout.isatty() if not in_pytest() else True
    if not output_to_stdout and not stdout_is_atty:
        print(entry)


def _make_file_name_banners(chains_to_filenames):
    same_width_names = _make_chain_code_to_names_same_width_names(chains_to_filenames)

    file_banners = {
        chain_code: f"------------- {same_width_names[chain_code]} -------------"
        for chain_code, file_name in chains_to_filenames.items()
    }
    banner_lengths = [len(banner) for banner in file_banners.values()]
    max_banner_width = max(banner_lengths)

    final_banner = "-" * max_banner_width

    return file_banners, final_banner


def _exit_if_chain_selector_not_in_chain_codes(chain_codes, chain_selectors):
    for chain_selector in chain_selectors:
        if chain_selector not in chain_codes:
            msg = f"""\
                the chain code {chain_selector} is not in the molecular systems chain codes
                chain selectors:  {', '.join(chain_selectors)}
                chain codes: {', '.join(chain_codes)}
            """
            exit_error(msg)


def _nef_to_nmrview_sequences(
    residues: List[SequenceResidue], chain_codes: List[str]
) -> Dict[str, List[str]]:

    chain_starts = get_chain_starts(residues)

    residues_by_chain = {
        chain_code: [
            residue for residue in residues if residue.chain_code == chain_code
        ]
        for chain_code in chain_codes
    }
    residue_sequences = {
        chain_code: [residue.residue_name for residue in residues_by_chain[chain_code]]
        for chain_code in chain_codes
    }
    result = {}
    for chain_code, residue_sequence in residue_sequences.items():
        chain_start = chain_starts[chain_code]

        sequence_lines = []
        for i, residue in enumerate(residue_sequence):

            residue = residue.lower()

            if i == 0:
                if chain_start != 1:
                    line = f"{residue} {chain_start}"
                else:
                    line = residue
            else:
                line = residue

            sequence_lines.append(line)

        result[chain_code] = "\n".join(sequence_lines)

    return result


def _nmrview_sequences_from_entry(entry, chain_codes):

    molecular_system_frame = molecular_system_from_entry(entry)

    residues = sequences_from_frames(molecular_system_frame)

    return _nef_to_nmrview_sequences(residues, chain_codes)


def _make_chain_code_to_names_same_width_names(names: Dict[str, str]) -> Dict[str, str]:
    max_length = -1
    for name in names.values():
        name_len = len(str(name))
        max_length = max(name_len, max_length)

    return {
        chain_code: str(name).center(max_length) for chain_code, name in names.items()
    }


def _open_for_writing_or_exit(file_name):
    try:
        result = open(file_name, "w")
    except Exception as e:
        msg = f"failed to open {file_name} for writing because {e}"

        exit_error(msg, e)

    return result


def _build_chains_to_filenames(
    chain_codes: List[str], file_name_template: str
) -> Dict[str, str]:

    return {chain_code: f(file_name_template) for chain_code in chain_codes}



import typer

from nef_pipelines import nef_app

simulate_app = typer.Typer()


if nef_app.app:
    nef_app.app.add_typer(simulate_app, name="simulate", help="- simulate data")

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.tools.simulate.peaks  # noqa: F401
    import nef_pipelines.tools.simulate.unlabelling  # noqa: F401



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.mars.exporters.shifts import shifts

app = typer.Typer()
app.command()(shifts)


def test_export_shifts_nef5_short():
    input = read_test_data("sec5_short.neff", __file__)

    result = run_and_report(app, "--out -".split(), input=input)

    EXPECTED = """\
              H      N        CA      CA-1    CB      CB-1
        PR_1  8.594  133.252  61.159  59.291  33.274  37.309
        PR_2  8.928  131.397  -       -       -       -
        PR_3  9.671  130.698  60.735  62.418  39.792  69.744
        PR_4  8.185  128.894  53.744  54.679  43.207  34.050
        PR_5  8.548  128.090  56.292  65.574  32.964  69.899
    """

    assert_lines_match(result.stdout, EXPECTED, squash_spaces=True)


def test_export_shifts_pete_co():

    input = read_test_data("short_co.neff", __file__)

    result = run_and_report(app, "--out -".split(), input=input)

    EXPECTED = """\
                   H        N  CA      CA-1    CB-1    CO       CO-1
        PR_1  11.671  136.682  -       -       -       -        -
        PR_2  10.946  131.689  -       -       -       -        -
        PR_3  10.266  112.518  46.281  63.862  32.984  176.722  178.726
        PR_4   9.632  104.529  46.465  -       -       173.390  175.767

    """

    assert_lines_match(result.stdout, EXPECTED, squash_spaces=True)


def test_export_shifts_david_bad_residue_name():

    input = read_test_data("sec5_short_bad_res_name.neff", __file__)

    result = run_and_report(
        app, "--out -".split(), input=input, expected_exit_code=1
    )  # , separate_stderr=True)

    # TODO: this should check in stderr not stdout but see note in test_lib: its currently not possible to capture
    # the stdout and stderr streams separately
    EXPECTED_ERROR_STRINGS = [
        "error for atom H in chain A",
        "the residue type in the sequence doesn't match residue type in shifts",
        "sequence residue name .",
        "frame name nef_chemical_shift_list_default",
        "row number 27",
    ]
    for error_string in EXPECTED_ERROR_STRINGS:
        assert error_string in result.stdout
